"""
If you use this code, please cite one of the SynthSeg papers:
https://github.com/BBillot/SynthSeg/blob/master/bibtex.bib

Copyright 2020 Benjamin Billot

Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software distributed under the License is
distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
implied. See the License for the specific language governing permissions and limitations under the
License.
"""

# python imports
import os
import csv
import numpy as np
import tensorflow as tf
import keras.layers as KL
import keras.backend as K
from keras.models import Model

# project imports
from lamareg.SynthSeg import evaluate

# third-party imports
from lamareg.ext.lab2im import utils
from lamareg.ext.lab2im import layers
from lamareg.ext.lab2im import edit_volumes
from lamareg.ext.neuron import models as nrn_models


def predict(
    path_images,
    path_segmentations,
    path_model,
    labels_segmentation,
    n_neutral_labels=None,
    names_segmentation=None,
    path_posteriors=None,
    path_resampled=None,
    path_volumes=None,
    min_pad=None,
    cropping=None,
    target_res=1.0,
    gradients=False,
    flip=True,
    topology_classes=None,
    sigma_smoothing=0.5,
    keep_biggest_component=True,
    n_levels=5,
    nb_conv_per_level=2,
    conv_size=3,
    unet_feat_count=24,
    feat_multiplier=2,
    activation="elu",
    gt_folder=None,
    evaluation_labels=None,
    list_incorrect_labels=None,
    list_correct_labels=None,
    compute_distances=False,
    recompute=True,
    verbose=True,
):
    """
    This function uses trained models to segment images.
    It is crucial that the inputs match the architecture parameters of the trained model.
    :param path_images: path of the images to segment. Can be the path to a directory or the path to a single image.
    :param path_segmentations: path where segmentations will be writen.
    Should be a dir, if path_images is a dir, and a file if path_images is a file.
    :param path_model: path ot the trained model.
    :param labels_segmentation: List of labels for which to compute Dice scores. It should be the same list as the
    segmentation_labels used in training.
    :param n_neutral_labels: (optional) if the label maps contain some right/left specific labels and if test-time
    flipping is applied (see parameter 'flip'), please provide the number of non-sided labels (including background).
    It should be the same value as for training. Default is None.
    :param names_segmentation: (optional) List of names corresponding to the names of the segmentation labels.
    Only used when path_volumes is provided. Must be of the same size as segmentation_labels. Can be given as a
    list, a numpy array of strings, or the path to such a numpy array. Default is None.
    :param path_posteriors: (optional) path where posteriors will be writen.
    Should be a dir, if path_images is a dir, and a file if path_images is a file.
    :param path_resampled: (optional) path where images resampled to 1mm isotropic will be writen.
    We emphasise that images are resampled as soon as the resolution in one of the axes is not in the range [0.9; 1.1].
    Should be a dir, if path_images is a dir, and a file if path_images is a file. Default is None, where resampled
    images are not saved.
    :param path_volumes: (optional) path of a csv file where the soft volumes of all segmented regions will be writen.
    The rows of the csv file correspond to subjects, and the columns correspond to segmentation labels.
    The soft volume of a structure corresponds to the sum of its predicted probability map.
    :param min_pad: (optional) minimum size of the images to process. Can be an int, a sequence or a 1d numpy array.
    :param cropping: (optional) crop the images to the specified shape before predicting the segmentation maps.
    Cropping overwrites min_pad if min_pad>cropping. Can be an int, a sequence or a 1d numpy array.
    :param target_res: (optional) target resolution at which the network operates (and thus resolution of the output
    segmentations). This must match the resolution of the training data ! target_res is used to automatically resampled
    the images with resolutions outside [target_res-0.05, target_res+0.05].
    :param gradients: (optional) whether to replace the image by the magnitude of its gradient as input to the network.
    Can be a sequence, a 1d numpy array. Set to None to disable the automatic resampling. Default is 1mm.
    :param flip: (optional) whether to perform test-time augmentation, where the input image is segmented along with
    a right/left flipped version on it. If set to True (default), be careful because this requires more memory.
    :param topology_classes: List of classes corresponding to all segmentation labels, in order to group them into
    classes, for each of which we will operate a smooth version of biggest connected component.
    Can be a sequence, a 1d numpy array, or the path to a numpy 1d array in the same order as segmentation_labels.
    Default is None, where no topological analysis is performed.
    :param sigma_smoothing: (optional) If not None, the posteriors are smoothed with a gaussian kernel of the specified
    standard deviation.
    :param keep_biggest_component: (optional) whether to only keep the biggest component in the predicted segmentation.
    This is applied independently of topology_classes, and it is applied to the whole segmentation
    :param n_levels: (optional) number of levels for unet. Default is 5.
    :param nb_conv_per_level: (optional) number of convolution layers per level. Default is 2.
    :param conv_size: (optional) size of UNet's convolution masks. Default is 3.
    :param unet_feat_count: (optional) number of features for the first layer of the unet. Default is 24.
    :param feat_multiplier: (optional) multiplicative factor for the number of feature for each new level. Default is 2.
    :param activation: (optional) activation function. Can be 'elu', 'relu'.
    :param gt_folder: (optional) path of the ground truth label maps corresponding to the input images. Should be a dir,
    if path_images is a dir, or a file if path_images is a file.
    Providing a gt_folder will trigger a Dice evaluation, where scores will be writen along with the path_segmentations.
    Specifically, the scores are contained in a numpy array, where labels are in rows, and subjects in columns.
    :param evaluation_labels: (optional) if gt_folder is True you can evaluate the Dice scores on a subset of the
    segmentation labels, by providing another label list here. Can be a sequence, a 1d numpy array, or the path to a
    numpy 1d array. Default is np.unique(segmentation_labels).
    :param list_incorrect_labels: (optional) this option enables to replace some label values in the obtained
    segmentations by other label values. Can be a list, a 1d numpy array, or the path to such an array.
    :param list_correct_labels: (optional) list of values to correct the labels specified in list_incorrect_labels.
    Correct values must have the same order as their corresponding value in list_incorrect_labels.
    :param compute_distances: (optional) whether to add Hausdorff and mean surface distance evaluations to the default
    Dice evaluation. Default is True.
    :param recompute: (optional) whether to recompute segmentations that were already computed. This also applies to
    Dice scores, if gt_folder is not None. Default is True.
    :param verbose: (optional) whether to print out info about the remaining number of cases.
    """

    # prepare input/output filepaths
    (
        path_images,
        path_segmentations,
        path_posteriors,
        path_resampled,
        path_volumes,
        compute,
        unique_vol_file,
    ) = prepare_output_files(
        path_images,
        path_segmentations,
        path_posteriors,
        path_resampled,
        path_volumes,
        recompute,
    )

    # get label list
    labels_segmentation, _ = utils.get_list_labels(label_list=labels_segmentation)
    if (n_neutral_labels is not None) & flip:
        labels_segmentation, flip_indices, unique_idx = get_flip_indices(
            labels_segmentation, n_neutral_labels
        )
    else:
        labels_segmentation, unique_idx = np.unique(
            labels_segmentation, return_index=True
        )
        flip_indices = None

    # prepare other labels list
    if names_segmentation is not None:
        names_segmentation = utils.load_array_if_path(names_segmentation)[unique_idx]
    if topology_classes is not None:
        topology_classes = utils.load_array_if_path(
            topology_classes, load_as_numpy=True
        )[unique_idx]

    # prepare volumes if necessary
    if unique_vol_file & (path_volumes[0] is not None):
        write_csv(path_volumes[0], None, True, labels_segmentation, names_segmentation)

    # build network
    _, _, n_dims, n_channels, _, _ = utils.get_volume_info(path_images[0])
    model_input_shape = [None] * n_dims + [n_channels]
    net = build_model(
        path_model=path_model,
        input_shape=model_input_shape,
        labels_segmentation=labels_segmentation,
        n_levels=n_levels,
        nb_conv_per_level=nb_conv_per_level,
        conv_size=conv_size,
        unet_feat_count=unet_feat_count,
        feat_multiplier=feat_multiplier,
        activation=activation,
        sigma_smoothing=sigma_smoothing,
        flip_indices=flip_indices,
        gradients=gradients,
    )

    # set cropping/padding
    if (cropping is not None) & (min_pad is not None):
        cropping = utils.reformat_to_list(cropping, length=n_dims, dtype="int")
        min_pad = utils.reformat_to_list(min_pad, length=n_dims, dtype="int")
        min_pad = np.minimum(cropping, min_pad)

    # perform segmentation
    if len(path_images) <= 10:
        loop_info = utils.LoopInfo(len(path_images), 1, "predicting", True)
    else:
        loop_info = utils.LoopInfo(len(path_images), 10, "predicting", True)
    for i in range(len(path_images)):
        if verbose:
            loop_info.update(i)

        # compute segmentation only if needed
        if compute[i]:

            # preprocessing
            image, aff, h, im_res, shape, pad_idx, crop_idx = preprocess(
                path_image=path_images[i],
                n_levels=n_levels,
                target_res=target_res,
                crop=cropping,
                min_pad=min_pad,
                path_resample=path_resampled[i],
            )

            # prediction
            post_patch = net.predict(image)

            # postprocessing
            seg, posteriors, volumes = postprocess(
                post_patch=post_patch,
                shape=shape,
                pad_idx=pad_idx,
                crop_idx=crop_idx,
                n_dims=n_dims,
                labels_segmentation=labels_segmentation,
                keep_biggest_component=keep_biggest_component,
                aff=aff,
                im_res=im_res,
                topology_classes=topology_classes,
            )

            # write results to disk
            utils.save_volume(seg, aff, h, path_segmentations[i], dtype="int32")
            if path_posteriors[i] is not None:
                if n_channels > 1:
                    posteriors = utils.add_axis(posteriors, axis=[0, -1])
                utils.save_volume(
                    posteriors, aff, h, path_posteriors[i], dtype="float32"
                )

            # compute volumes
            if path_volumes[i] is not None:
                row = [os.path.basename(path_images[i]).replace(".nii.gz", "")] + [
                    str(vol) for vol in volumes
                ]
                write_csv(
                    path_volumes[i],
                    row,
                    unique_vol_file,
                    labels_segmentation,
                    names_segmentation,
                )

    # evaluate
    if gt_folder is not None:

        # find path where segmentations are saved evaluation folder, and get labels on which to evaluate
        eval_folder = os.path.dirname(path_segmentations[0])
        if evaluation_labels is None:
            evaluation_labels = labels_segmentation

        # set path of result arrays for surface distance if necessary
        if compute_distances:
            path_hausdorff = os.path.join(eval_folder, "hausdorff.npy")
            path_hausdorff_99 = os.path.join(eval_folder, "hausdorff_99.npy")
            path_hausdorff_95 = os.path.join(eval_folder, "hausdorff_95.npy")
            path_mean_distance = os.path.join(eval_folder, "mean_distance.npy")
        else:
            path_hausdorff = path_hausdorff_99 = path_hausdorff_95 = (
                path_mean_distance
            ) = None

        # compute evaluation metrics
        evaluate.evaluation(
            gt_folder,
            eval_folder,
            evaluation_labels,
            path_dice=os.path.join(eval_folder, "dice.npy"),
            path_hausdorff=path_hausdorff,
            path_hausdorff_99=path_hausdorff_99,
            path_hausdorff_95=path_hausdorff_95,
            path_mean_distance=path_mean_distance,
            list_incorrect_labels=list_incorrect_labels,
            list_correct_labels=list_correct_labels,
            recompute=recompute,
            verbose=verbose,
        )


def prepare_output_files(
    path_images, out_seg, out_posteriors, out_resampled, out_volumes, recompute
):

    # check inputs
    assert path_images is not None, "please specify an input file/folder (--i)"
    assert out_seg is not None, "please specify an output file/folder (--o)"

    # convert path to absolute paths
    path_images = os.path.abspath(path_images)
    basename = os.path.basename(path_images)
    out_seg = os.path.abspath(out_seg)
    out_posteriors = (
        os.path.abspath(out_posteriors)
        if (out_posteriors is not None)
        else out_posteriors
    )
    out_resampled = (
        os.path.abspath(out_resampled) if (out_resampled is not None) else out_resampled
    )
    out_volumes = (
        os.path.abspath(out_volumes) if (out_volumes is not None) else out_volumes
    )

    # path_images is a text file
    if basename[-4:] == ".txt":

        # input images
        if not os.path.isfile(path_images):
            raise Exception(
                "provided text file containing paths of input images does not exist"
                % path_images
            )
        with open(path_images, "r") as f:
            path_images = [
                line.replace("\n", "") for line in f.readlines() if line != "\n"
            ]

        # define helper to deal with outputs
        def text_helper(path, name):
            if path is not None:
                assert path[-4:] == ".txt", (
                    "if path_images given as text file, so must be %s" % name
                )
                with open(path, "r") as ff:
                    path = [
                        line.replace("\n", "")
                        for line in ff.readlines()
                        if line != "\n"
                    ]
                recompute_files = [not os.path.isfile(p) for p in path]
            else:
                path = [None] * len(path_images)
                recompute_files = [False] * len(path_images)
            unique_file = False
            return path, recompute_files, unique_file

        # use helper on all outputs
        out_seg, recompute_seg, _ = text_helper(out_seg, "path_segmentations")
        out_posteriors, recompute_post, _ = text_helper(
            out_posteriors, "path_posteriors"
        )
        out_resampled, recompute_resampled, _ = text_helper(
            out_resampled, "path_resampled"
        )
        out_volumes, recompute_volume, unique_volume_file = text_helper(
            out_volumes, "path_volume"
        )

    # path_images is a folder
    elif (
        (".nii.gz" not in basename)
        & (".nii" not in basename)
        & (".mgz" not in basename)
        & (".npz" not in basename)
    ):

        # input images
        if os.path.isfile(path_images):
            raise Exception(
                "Extension not supported for %s, only use: nii.gz, .nii, .mgz, or .npz"
                % path_images
            )
        path_images = utils.list_images_in_folder(path_images)

        # define helper to deal with outputs
        def helper_dir(path, name, file_type, suffix):
            unique_file = False
            if path is not None:
                assert path[-4:] != ".txt", (
                    "%s can only be given as text file when path_images is." % name
                )
                if file_type == "csv":
                    if path[-4:] != ".csv":
                        print(
                            "%s provided without csv extension. Adding csv extension."
                            % name
                        )
                        path += ".csv"
                    path = [path] * len(path_images)
                    recompute_files = [True] * len(path_images)
                    unique_file = True
                else:
                    if (
                        (path[-7:] == ".nii.gz")
                        | (path[-4:] == ".nii")
                        | (path[-4:] == ".mgz")
                        | (path[-4:] == ".npz")
                    ):
                        raise Exception("Output FOLDER had a FILE extension" % path)
                    path = [
                        os.path.join(path, os.path.basename(p)) for p in path_images
                    ]
                    path = [p.replace(".nii", "_%s.nii" % suffix) for p in path]
                    path = [p.replace(".mgz", "_%s.mgz" % suffix) for p in path]
                    path = [p.replace(".npz", "_%s.npz" % suffix) for p in path]
                    recompute_files = [not os.path.isfile(p) for p in path]
                utils.mkdir(os.path.dirname(path[0]))
            else:
                path = [None] * len(path_images)
                recompute_files = [False] * len(path_images)
            return path, recompute_files, unique_file

        # use helper on all outputs
        out_seg, recompute_seg, _ = helper_dir(
            out_seg, "path_segmentations", "", "synthseg"
        )
        out_posteriors, recompute_post, _ = helper_dir(
            out_posteriors, "path_posteriors", "", "posteriors"
        )
        out_resampled, recompute_resampled, _ = helper_dir(
            out_resampled, "path_resampled", "", "resampled"
        )
        out_volumes, recompute_volume, unique_volume_file = helper_dir(
            out_volumes, "path_volumes", "csv", ""
        )

    # path_images is an image
    else:

        # input image
        assert os.path.isfile(path_images), (
            "file does not exist: %s \n"
            "please make sure the path and the extension are correct" % path_images
        )
        path_images = [path_images]

        # define helper to deal with outputs
        def helper_im(path, name, file_type, suffix):
            unique_file = False
            if path is not None:
                assert path[-4:] != ".txt", (
                    "%s can only be given as text file when path_images is." % name
                )
                if file_type == "csv":
                    if path[-4:] != ".csv":
                        print(
                            "%s provided without csv extension. Adding csv extension."
                            % name
                        )
                        path += ".csv"
                    recompute_files = [True]
                    unique_file = True
                else:
                    if (
                        (".nii.gz" not in path)
                        & (".nii" not in path)
                        & (".mgz" not in path)
                        & (".npz" not in path)
                    ):
                        file_name = os.path.basename(path_images[0]).replace(
                            ".nii", "_%s.nii" % suffix
                        )
                        file_name = file_name.replace(".mgz", "_%s.mgz" % suffix)
                        file_name = file_name.replace(".npz", "_%s.npz" % suffix)
                        path = os.path.join(path, file_name)
                    recompute_files = [not os.path.isfile(path)]
                utils.mkdir(os.path.dirname(path))
            else:
                recompute_files = [False]
            path = [path]
            return path, recompute_files, unique_file

        # use helper on all outputs
        out_seg, recompute_seg, _ = helper_im(
            out_seg, "path_segmentations", "", "synthseg"
        )
        out_posteriors, recompute_post, _ = helper_im(
            out_posteriors, "path_posteriors", "", "posteriors"
        )
        out_resampled, recompute_resampled, _ = helper_im(
            out_resampled, "path_resampled", "", "resampled"
        )
        out_volumes, recompute_volume, unique_volume_file = helper_im(
            out_volumes, "path_volumes", "csv", ""
        )

    recompute_list = [
        recompute | re_seg | re_post | re_res | re_vol
        for (re_seg, re_post, re_res, re_vol) in zip(
            recompute_seg, recompute_post, recompute_resampled, recompute_volume
        )
    ]

    return (
        path_images,
        out_seg,
        out_posteriors,
        out_resampled,
        out_volumes,
        recompute_list,
        unique_volume_file,
    )


def preprocess(
    path_image, n_levels, target_res, crop=None, min_pad=None, path_resample=None
):

    # read image and corresponding info
    im, _, aff, n_dims, n_channels, h, im_res = utils.get_volume_info(path_image, True)

    # resample image if necessary
    if target_res is not None:
        target_res = np.squeeze(utils.reformat_to_n_channels_array(target_res, n_dims))
        if np.any((im_res > target_res + 0.05) | (im_res < target_res - 0.05)):
            im_res = target_res
            im, aff = edit_volumes.resample_volume(im, aff, im_res)
            if path_resample is not None:
                utils.save_volume(im, aff, h, path_resample)

    # align image
    im = edit_volumes.align_volume_to_ref(
        im, aff, aff_ref=np.eye(4), n_dims=n_dims, return_copy=False
    )
    shape = list(im.shape[:n_dims])

    # crop image if necessary
    if crop is not None:
        crop = utils.reformat_to_list(crop, length=n_dims, dtype="int")
        crop_shape = [
            utils.find_closest_number_divisible_by_m(s, 2**n_levels, "higher")
            for s in crop
        ]
        im, crop_idx = edit_volumes.crop_volume(
            im, cropping_shape=crop_shape, return_crop_idx=True
        )
    else:
        crop_idx = None

    # normalise image
    if n_channels == 1:
        im = edit_volumes.rescale_volume(
            im, new_min=0.0, new_max=1.0, min_percentile=0.5, max_percentile=99.5
        )
    else:
        for i in range(im.shape[-1]):
            im[..., i] = edit_volumes.rescale_volume(
                im[..., i],
                new_min=0.0,
                new_max=1.0,
                min_percentile=0.5,
                max_percentile=99.5,
            )

    # pad image
    input_shape = im.shape[:n_dims]
    pad_shape = [
        utils.find_closest_number_divisible_by_m(s, 2**n_levels, "higher")
        for s in input_shape
    ]
    if (
        min_pad is not None
    ):  # in SynthSeg predict use crop flag and then if used do min_pad=crop else min_pad = 192
        min_pad = utils.reformat_to_list(min_pad, length=n_dims, dtype="int")
        min_pad = [
            utils.find_closest_number_divisible_by_m(s, 2**n_levels, "higher")
            for s in min_pad
        ]
        pad_shape = np.maximum(pad_shape, min_pad)
    im, pad_idx = edit_volumes.pad_volume(
        im, padding_shape=pad_shape, return_pad_idx=True
    )

    # add batch and channel axes
    im = utils.add_axis(im) if n_channels > 1 else utils.add_axis(im, axis=[0, -1])

    return im, aff, h, im_res, shape, pad_idx, crop_idx


def build_model(
    path_model,
    input_shape,
    labels_segmentation,
    n_levels,
    nb_conv_per_level,
    conv_size,
    unet_feat_count,
    feat_multiplier,
    activation,
    sigma_smoothing,
    flip_indices,
    gradients,
):

    assert os.path.isfile(path_model), "The provided model path does not exist."

    # get labels
    n_labels_seg = len(labels_segmentation)

    if gradients:
        input_image = KL.Input(input_shape)
        last_tensor = layers.ImageGradients("sobel", True)(input_image)
        last_tensor = KL.Lambda(
            lambda x: (x - K.min(x)) / (K.max(x) - K.min(x) + K.epsilon())
        )(last_tensor)
        net = Model(inputs=input_image, outputs=last_tensor)
    else:
        net = None

    # build UNet
    net = nrn_models.unet(
        input_model=net,
        input_shape=input_shape,
        nb_labels=n_labels_seg,
        nb_levels=n_levels,
        nb_conv_per_level=nb_conv_per_level,
        conv_size=conv_size,
        nb_features=unet_feat_count,
        feat_mult=feat_multiplier,
        activation=activation,
        batch_norm=-1,
    )
    net.load_weights(path_model, by_name=True)

    # smooth posteriors if specified
    if sigma_smoothing > 0:
        last_tensor = net.output
        last_tensor._keras_shape = tuple(last_tensor.get_shape().as_list())
        last_tensor = layers.GaussianBlur(sigma=sigma_smoothing)(last_tensor)
        net = Model(inputs=net.inputs, outputs=last_tensor)

    if flip_indices is not None:

        # segment flipped image
        input_image = net.inputs[0]
        seg = net.output
        image_flipped = layers.RandomFlip(axis=0, prob=1)(input_image)
        last_tensor = net(image_flipped)

        # flip back and re-order channels
        last_tensor = layers.RandomFlip(axis=0, prob=1)(last_tensor)
        last_tensor = KL.Lambda(
            lambda x: tf.split(x, [1] * n_labels_seg, axis=-1), name="split"
        )(last_tensor)
        reordered_channels = [last_tensor[flip_indices[i]] for i in range(n_labels_seg)]
        last_tensor = KL.Lambda(lambda x: tf.concat(x, -1), name="concat")(
            reordered_channels
        )

        # average two segmentations and build model
        name_segm_prediction_layer = "average_lr"
        last_tensor = KL.Lambda(
            lambda x: 0.5 * (x[0] + x[1]), name=name_segm_prediction_layer
        )([seg, last_tensor])
        net = Model(inputs=net.inputs, outputs=last_tensor)

    return net


def postprocess(
    post_patch,
    shape,
    pad_idx,
    crop_idx,
    n_dims,
    labels_segmentation,
    keep_biggest_component,
    aff,
    im_res,
    topology_classes=None,
):

    # get posteriors
    post_patch = np.squeeze(post_patch)
    if topology_classes is None:
        post_patch = edit_volumes.crop_volume_with_idx(
            post_patch, pad_idx, n_dims=3, return_copy=False
        )

    # keep biggest connected component
    if keep_biggest_component:
        tmp_post_patch = post_patch[..., 1:]
        post_patch_mask = np.sum(tmp_post_patch, axis=-1) > 0.25
        post_patch_mask = edit_volumes.get_largest_connected_component(post_patch_mask)
        post_patch_mask = np.stack(
            [post_patch_mask] * tmp_post_patch.shape[-1], axis=-1
        )
        tmp_post_patch = edit_volumes.mask_volume(
            tmp_post_patch, mask=post_patch_mask, return_copy=False
        )
        post_patch[..., 1:] = tmp_post_patch

    # reset posteriors to zero outside the largest connected component of each topological class
    if topology_classes is not None:
        post_patch_mask = post_patch > 0.25
        for topology_class in np.unique(topology_classes)[1:]:
            tmp_topology_indices = np.where(topology_classes == topology_class)[0]
            tmp_mask = np.any(post_patch_mask[..., tmp_topology_indices], axis=-1)
            tmp_mask = edit_volumes.get_largest_connected_component(tmp_mask)
            for idx in tmp_topology_indices:
                post_patch[..., idx] *= tmp_mask
        post_patch = edit_volumes.crop_volume_with_idx(
            post_patch, pad_idx, n_dims=3, return_copy=False
        )

    # normalise posteriors and get hard segmentation
    if keep_biggest_component | (topology_classes is not None):
        post_patch /= np.sum(post_patch, axis=-1)[..., np.newaxis]
    seg_patch = labels_segmentation[post_patch.argmax(-1).astype("int32")].astype(
        "int32"
    )

    # paste patches back to matrix of original image size
    if crop_idx is not None:
        # we need to go through this because of the posteriors of the background, otherwise pad_volume would work
        seg = np.zeros(shape=shape, dtype="int32")
        posteriors = np.zeros(shape=[*shape, labels_segmentation.shape[0]])
        posteriors[..., 0] = np.ones(shape)  # place background around patch
        if n_dims == 2:
            seg[crop_idx[0] : crop_idx[2], crop_idx[1] : crop_idx[3]] = seg_patch
            posteriors[crop_idx[0] : crop_idx[2], crop_idx[1] : crop_idx[3], :] = (
                post_patch
            )
        elif n_dims == 3:
            seg[
                crop_idx[0] : crop_idx[3],
                crop_idx[1] : crop_idx[4],
                crop_idx[2] : crop_idx[5],
            ] = seg_patch
            posteriors[
                crop_idx[0] : crop_idx[3],
                crop_idx[1] : crop_idx[4],
                crop_idx[2] : crop_idx[5],
                :,
            ] = post_patch
    else:
        seg = seg_patch
        posteriors = post_patch

    # align prediction back to first orientation
    seg = edit_volumes.align_volume_to_ref(
        seg, aff=np.eye(4), aff_ref=aff, n_dims=n_dims, return_copy=False
    )
    posteriors = edit_volumes.align_volume_to_ref(
        posteriors, np.eye(4), aff_ref=aff, n_dims=n_dims, return_copy=False
    )

    # compute volumes
    volumes = np.sum(
        posteriors[..., 1:], axis=tuple(range(0, len(posteriors.shape) - 1))
    )
    volumes = np.around(volumes * np.prod(im_res), 3)

    return seg, posteriors, volumes


def get_flip_indices(labels_segmentation, n_neutral_labels):

    # get position labels
    n_sided_labels = int((len(labels_segmentation) - n_neutral_labels) / 2)
    neutral_labels = labels_segmentation[:n_neutral_labels]
    left = labels_segmentation[n_neutral_labels : n_neutral_labels + n_sided_labels]

    # get correspondence between labels
    lr_corresp = np.stack(
        [
            labels_segmentation[n_neutral_labels : n_neutral_labels + n_sided_labels],
            labels_segmentation[n_neutral_labels + n_sided_labels :],
        ]
    )
    lr_corresp_unique, lr_corresp_indices = np.unique(
        lr_corresp[0, :], return_index=True
    )
    lr_corresp_unique = np.stack([lr_corresp_unique, lr_corresp[1, lr_corresp_indices]])
    lr_corresp_unique = (
        lr_corresp_unique[:, 1:]
        if not np.all(lr_corresp_unique[:, 0])
        else lr_corresp_unique
    )

    # get unique labels
    labels_segmentation, unique_idx = np.unique(labels_segmentation, return_index=True)

    # get indices of corresponding labels
    lr_indices = np.zeros_like(lr_corresp_unique)
    for i in range(lr_corresp_unique.shape[0]):
        for j, lab in enumerate(lr_corresp_unique[i]):
            lr_indices[i, j] = np.where(labels_segmentation == lab)[0]

    # build 1d vector to swap LR corresponding labels taking into account neutral labels
    flip_indices = np.zeros_like(labels_segmentation)
    for i in range(len(flip_indices)):
        if labels_segmentation[i] in neutral_labels:
            flip_indices[i] = i
        elif labels_segmentation[i] in left:
            flip_indices[i] = lr_indices[
                1, np.where(lr_corresp_unique[0, :] == labels_segmentation[i])
            ]
        else:
            flip_indices[i] = lr_indices[
                0, np.where(lr_corresp_unique[1, :] == labels_segmentation[i])
            ]

    return labels_segmentation, flip_indices, unique_idx


def write_csv(
    path_csv, data, unique_file, labels, names, skip_first=True, last_first=False
):

    # initialisation
    utils.mkdir(os.path.dirname(path_csv))
    labels, unique_idx = np.unique(labels, return_index=True)
    if skip_first:
        labels = labels[1:]
    if names is not None:
        names = names[unique_idx].tolist()
        if skip_first:
            names = names[1:]
        header = names
    else:
        header = [str(lab) for lab in labels]
    if last_first:
        header = [header[-1]] + header[:-1]
    if (not unique_file) & (data is None):
        raise ValueError("data can only be None when initialising a unique volume file")

    # modify data
    if unique_file:
        if data is None:
            type_open = "w"
            data = ["subject"] + header
        else:
            type_open = "a"
        data = [data]
    else:
        type_open = "w"
        header = [""] + header
        data = [header, data]

    # write csv
    with open(path_csv, type_open) as csvFile:
        writer = csv.writer(csvFile)
        writer.writerows(data)



from . import brain_generator
from . import estimate_priors
from . import evaluate
from . import labels_to_image_model
from . import metrics_model
from . import model_inputs
from . import predict
from . import training_supervised
from . import training



"""
If you use this code, please cite one of the SynthSeg papers:
https://github.com/BBillot/SynthSeg/blob/master/bibtex.bib

Copyright 2020 Benjamin Billot

Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software distributed under the License is
distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
implied. See the License for the specific language governing permissions and limitations under the
License.
"""

# python imports
import os
import sys
import traceback
import numpy as np
import tensorflow as tf
import keras.layers as KL
import keras.backend as K
from keras.models import Model

# project imports
from lamareg.SynthSeg import evaluate
from lamareg.SynthSeg.predict import write_csv, get_flip_indices
from lamareg.SynthSeg.model_handler import ensure_model_file_exists

# third-party imports
from lamareg.ext.lab2im import utils
from lamareg.ext.lab2im import layers
from lamareg.ext.lab2im import edit_volumes
from lamareg.ext.neuron import models as nrn_models


def predict(
    path_images,
    path_segmentations,
    path_model_segmentation,
    labels_segmentation,
    robust,
    fast,
    v1,
    n_neutral_labels,
    labels_denoiser,
    path_posteriors,
    path_resampled,
    path_volumes,
    do_parcellation,
    path_model_parcellation,
    labels_parcellation,
    path_qc_scores,
    path_model_qc,
    labels_qc,
    cropping,
    ct=False,
    names_segmentation=None,
    names_parcellation=None,
    names_qc=None,
    topology_classes=None,
    sigma_smoothing=0.5,
    input_shape_qc=224,
    gt_folder=None,
    evaluation_labels=None,
    mask_folder=None,
    list_incorrect_labels=None,
    list_correct_labels=None,
    compute_distances=False,
    recompute=True,
    verbose=True,
):

    # prepare input/output filepaths
    outputs = prepare_output_files(
        path_images,
        path_segmentations,
        path_posteriors,
        path_resampled,
        path_volumes,
        path_qc_scores,
        recompute,
    )
    path_images = outputs[0]
    path_segmentations = outputs[1]
    path_posteriors = outputs[2]
    path_resampled = outputs[3]
    path_volumes = outputs[4]
    unique_vol_file = outputs[5]
    path_qc_scores = outputs[6]
    unique_qc_file = outputs[7]
    compute = outputs[8]

    # get label lists
    labels_segmentation, _ = utils.get_list_labels(label_list=labels_segmentation)
    if (n_neutral_labels is not None) & (not fast) & (not robust):
        labels_segmentation, flip_indices, unique_idx = get_flip_indices(
            labels_segmentation, n_neutral_labels
        )
    else:
        labels_segmentation, unique_idx = np.unique(
            labels_segmentation, return_index=True
        )
        flip_indices = None

    # prepare other labels list
    if names_segmentation is not None:
        names_segmentation = utils.load_array_if_path(names_segmentation)[unique_idx]
    if topology_classes is not None:
        topology_classes = utils.load_array_if_path(
            topology_classes, load_as_numpy=True
        )[unique_idx]
    labels_denoiser = np.unique(utils.get_list_labels(labels_denoiser)[0])
    if do_parcellation:
        labels_parcellation, unique_i_parc = np.unique(
            utils.get_list_labels(labels_parcellation)[0], return_index=True
        )
        labels_volumes = np.concatenate([labels_segmentation, labels_parcellation[1:]])
        if (names_parcellation is not None) & (names_segmentation is not None):
            names_parcellation = utils.load_array_if_path(names_parcellation)[
                unique_i_parc
            ][1:]
            names_volumes = np.concatenate([names_segmentation, names_parcellation])
        else:
            names_volumes = names_segmentation
    else:
        labels_volumes = labels_segmentation
        names_volumes = names_segmentation
    if not v1:
        labels_volumes = np.concatenate(
            [labels_volumes, np.array([np.max(labels_volumes + 1)])]
        )
        if names_segmentation is not None:
            names_volumes = np.concatenate(
                [names_volumes, np.array(["total intracranial"])]
            )
    do_qc = True if path_qc_scores[0] is not None else False
    if do_qc:
        labels_qc = utils.get_list_labels(labels_qc)[0][unique_idx]
        if names_qc is not None:
            names_qc = utils.load_array_if_path(names_qc)[unique_idx]

    # prepare volume/QC files if necessary
    if unique_vol_file & (path_volumes[0] is not None):
        write_csv(
            path_volumes[0],
            None,
            True,
            labels_volumes,
            names_volumes,
            last_first=(not v1),
        )
    if unique_qc_file & do_qc:
        write_csv(path_qc_scores[0], None, True, labels_qc, names_qc)

    # build network
    net = build_model(
        path_model_segmentation=path_model_segmentation,
        path_model_parcellation=path_model_parcellation,
        path_model_qc=path_model_qc,
        input_shape_qc=input_shape_qc,
        labels_segmentation=labels_segmentation,
        labels_denoiser=labels_denoiser,
        labels_parcellation=labels_parcellation,
        labels_qc=labels_qc,
        sigma_smoothing=sigma_smoothing,
        flip_indices=flip_indices,
        robust=robust,
        do_parcellation=do_parcellation,
        do_qc=do_qc,
    )

    # set cropping/padding
    if cropping is not None:
        cropping = utils.reformat_to_list(cropping, length=3, dtype="int")
        min_pad = cropping
    else:
        min_pad = 128

    # perform segmentation
    if len(path_images) <= 10:
        loop_info = utils.LoopInfo(len(path_images), 1, "predicting", True)
    else:
        loop_info = utils.LoopInfo(len(path_images), 10, "predicting", True)
    list_errors = list()
    for i in range(len(path_images)):
        if verbose:
            loop_info.update(i)

        # compute segmentation only if needed
        if compute[i]:

            try:

                # preprocessing
                image, aff, h, im_res, shape, pad_idx, crop_idx = preprocess(
                    path_image=path_images[i],
                    ct=ct,
                    crop=cropping,
                    min_pad=min_pad,
                    path_resample=path_resampled[i],
                )

                # prediction
                shape_input = utils.add_axis(np.array(image.shape[1:-1]))
                if do_parcellation & do_qc:
                    post_patch_segmentation, post_patch_parcellation, qc_score = (
                        net.predict([image, shape_input])
                    )
                elif do_parcellation & (not do_qc):
                    post_patch_segmentation, post_patch_parcellation = net.predict(
                        image
                    )
                    qc_score = None
                elif (not do_parcellation) & do_qc:
                    post_patch_segmentation, qc_score = net.predict(
                        [image, shape_input]
                    )
                    post_patch_parcellation = None
                else:
                    post_patch_segmentation = net.predict(image)
                    post_patch_parcellation = qc_score = None

                # postprocessing
                seg, posteriors, volumes = postprocess(
                    post_patch_seg=post_patch_segmentation,
                    post_patch_parc=post_patch_parcellation,
                    shape=shape,
                    pad_idx=pad_idx,
                    crop_idx=crop_idx,
                    labels_segmentation=labels_segmentation,
                    labels_parcellation=labels_parcellation,
                    aff=aff,
                    im_res=im_res,
                    fast=fast,
                    topology_classes=topology_classes,
                    v1=v1,
                )

                # write predictions to disc
                utils.save_volume(seg, aff, h, path_segmentations[i], dtype="int32")
                if path_posteriors[i] is not None:
                    utils.save_volume(
                        posteriors, aff, h, path_posteriors[i], dtype="float32"
                    )

                # write volumes to disc if necessary
                if path_volumes[i] is not None:
                    row = [os.path.basename(path_images[i]).replace(".nii.gz", "")] + [
                        str(vol) for vol in volumes
                    ]
                    write_csv(
                        path_volumes[i],
                        row,
                        unique_vol_file,
                        labels_volumes,
                        names_volumes,
                        last_first=(not v1),
                    )

                # write QC scores to disc if necessary
                if path_qc_scores[i] is not None:
                    qc_score = np.around(np.clip(np.squeeze(qc_score)[1:], 0, 1), 4)
                    row = [os.path.basename(path_images[i]).replace(".nii.gz", "")] + [
                        "%.4f" % q for q in qc_score
                    ]
                    write_csv(
                        path_qc_scores[i], row, unique_qc_file, labels_qc, names_qc
                    )

            except Exception as e:
                list_errors.append(path_images[i])
                print(
                    "\nthe following problem occurred with image %s :" % path_images[i]
                )
                print(traceback.format_exc())
                print("resuming program execution\n")
                continue

    # print output info
    if len(path_segmentations) == 1:  # only one image is processed
        print("\nsegmentation  saved in:    " + path_segmentations[0])
        if path_posteriors[0] is not None:
            print("posteriors saved in:       " + path_posteriors[0])
        if path_resampled[0] is not None:
            print("resampled image saved in:  " + path_resampled[0])
        if path_volumes[0] is not None:
            print("volumes saved in:          " + path_volumes[0])
        if path_qc_scores[0] is not None:
            print("QC scores saved in:        " + path_qc_scores[0])
    else:  # all segmentations are in the same folder, and we have unique vol/QC files
        if (
            len(
                set(
                    [
                        os.path.dirname(path_segmentations[i])
                        for i in range(len(path_segmentations))
                    ]
                )
            )
            <= 1
        ):
            print(
                "\nsegmentations saved in:    " + os.path.dirname(path_segmentations[0])
            )
            if path_posteriors[0] is not None:
                print(
                    "posteriors saved in:       " + os.path.dirname(path_posteriors[0])
                )
            if path_resampled[0] is not None:
                print(
                    "resampled images saved in: " + os.path.dirname(path_resampled[0])
                )
            if path_volumes[0] is not None:
                print("volumes saved in:          " + path_volumes[0])
            if path_qc_scores[0] is not None:
                print("QC scores saved in:        " + path_qc_scores[0])

    if robust:
        print(
            "\nIf you use the new robust version of SynthSeg in a publication, please cite:"
        )
        print(
            "Robust machine learning segmentation for large-scale analysis of heterogeneous clinical brain MRI "
            "datasets"
        )
        print("B. Billot, M. Collin, S.E. Arnold, S. Das, J.E. Iglesias")
    else:
        print("\nIf you use this tool in a publication, please cite:")
        print(
            "SynthSeg: domain randomisation for segmentation of brain MRI scans of any contrast and resolution"
        )
        print(
            "B. Billot, D.N. Greve, O. Puonti, A. Thielscher, K. Van Leemput, B. Fischl, A.V. Dalca, J.E. Iglesias"
        )

    if len(list_errors) > 0:
        print(
            "\nERROR: some problems occurred for the following inputs (see corresponding errors above):"
        )
        for path_error_image in list_errors:
            print(path_error_image)
        sys.exit(1)

    # evaluate
    if gt_folder is not None:

        # find path where segmentations are saved evaluation folder, and get labels on which to evaluate
        eval_folder = os.path.dirname(path_segmentations[0])
        if evaluation_labels is None:
            evaluation_labels = labels_segmentation

        # set path of result arrays for surface distance if necessary
        if compute_distances:
            path_hausdorff = os.path.join(eval_folder, "hausdorff.npy")
            path_hausdorff_99 = os.path.join(eval_folder, "hausdorff_99.npy")
            path_hausdorff_95 = os.path.join(eval_folder, "hausdorff_95.npy")
            path_mean_distance = os.path.join(eval_folder, "mean_distance.npy")
        else:
            path_hausdorff = path_hausdorff_99 = path_hausdorff_95 = (
                path_mean_distance
            ) = None

        # compute evaluation metrics
        evaluate.evaluation(
            gt_folder,
            eval_folder,
            evaluation_labels,
            mask_dir=mask_folder,
            path_dice=os.path.join(eval_folder, "dice.npy"),
            path_hausdorff=path_hausdorff,
            path_hausdorff_99=path_hausdorff_99,
            path_hausdorff_95=path_hausdorff_95,
            path_mean_distance=path_mean_distance,
            list_incorrect_labels=list_incorrect_labels,
            list_correct_labels=list_correct_labels,
            recompute=recompute,
            verbose=verbose,
        )


def prepare_output_files(
    path_images, out_seg, out_posteriors, out_resampled, out_volumes, out_qc, recompute
):

    # check inputs
    assert path_images is not None, "please specify an input file/folder (--i)"
    assert out_seg is not None, "please specify an output file/folder (--o)"

    # convert path to absolute paths
    path_images = os.path.abspath(path_images)
    basename = os.path.basename(path_images)
    out_seg = os.path.abspath(out_seg)
    out_posteriors = (
        os.path.abspath(out_posteriors)
        if (out_posteriors is not None)
        else out_posteriors
    )
    out_resampled = (
        os.path.abspath(out_resampled) if (out_resampled is not None) else out_resampled
    )
    out_volumes = (
        os.path.abspath(out_volumes) if (out_volumes is not None) else out_volumes
    )
    out_qc = os.path.abspath(out_qc) if (out_qc is not None) else out_qc

    # path_images is a text file
    if basename[-4:] == ".txt":

        # input images
        if not os.path.isfile(path_images):
            raise Exception(
                "provided text file containing paths of input images does not exist"
                % path_images
            )
        with open(path_images, "r") as f:
            path_images = [
                line.replace("\n", "") for line in f.readlines() if line != "\n"
            ]

        # define helper to deal with outputs
        def text_helper(path, name):
            if path is not None:
                assert path[-4:] == ".txt", (
                    "if path_images given as text file, so must be %s" % name
                )
                with open(path, "r") as ff:
                    path = [
                        line.replace("\n", "")
                        for line in ff.readlines()
                        if line != "\n"
                    ]
                recompute_files = [not os.path.isfile(p) for p in path]
            else:
                path = [None] * len(path_images)
                recompute_files = [False] * len(path_images)
            unique_file = False
            return path, recompute_files, unique_file

        # use helper on all outputs
        out_seg, recompute_seg, _ = text_helper(out_seg, "path_segmentations")
        out_posteriors, recompute_post, _ = text_helper(
            out_posteriors, "path_posteriors"
        )
        out_resampled, recompute_resampled, _ = text_helper(
            out_resampled, "path_resampled"
        )
        out_volumes, recompute_volume, unique_volume_file = text_helper(
            out_volumes, "path_volume"
        )
        out_qc, recompute_qc, unique_qc_file = text_helper(out_qc, "path_qc_scores")

    # path_images is a folder
    elif (
        (".nii.gz" not in basename)
        & (".nii" not in basename)
        & (".mgz" not in basename)
        & (".npz" not in basename)
    ):

        # input images
        if os.path.isfile(path_images):
            raise Exception(
                "Extension not supported for %s, only use: nii.gz, .nii, .mgz, or .npz"
                % path_images
            )
        path_images = utils.list_images_in_folder(path_images)

        # define helper to deal with outputs
        def helper_dir(path, name, file_type, suffix):
            unique_file = False
            if path is not None:
                assert path[-4:] != ".txt", (
                    "%s can only be given as text file when path_images is." % name
                )
                if file_type == "csv":
                    if path[-4:] != ".csv":
                        print(
                            "%s provided without csv extension. Adding csv extension."
                            % name
                        )
                        path += ".csv"
                    path = [path] * len(path_images)
                    recompute_files = [True] * len(path_images)
                    unique_file = True
                else:
                    if (
                        (path[-7:] == ".nii.gz")
                        | (path[-4:] == ".nii")
                        | (path[-4:] == ".mgz")
                        | (path[-4:] == ".npz")
                    ):
                        raise Exception("Output FOLDER had a FILE extension" % path)
                    path = [
                        os.path.join(path, os.path.basename(p)) for p in path_images
                    ]
                    path = [p.replace(".nii", "_%s.nii" % suffix) for p in path]
                    path = [p.replace(".mgz", "_%s.mgz" % suffix) for p in path]
                    path = [p.replace(".npz", "_%s.npz" % suffix) for p in path]
                    recompute_files = [not os.path.isfile(p) for p in path]
                utils.mkdir(os.path.dirname(path[0]))
            else:
                path = [None] * len(path_images)
                recompute_files = [False] * len(path_images)
            return path, recompute_files, unique_file

        # use helper on all outputs
        out_seg, recompute_seg, _ = helper_dir(
            out_seg, "path_segmentations", "", "synthseg"
        )
        out_posteriors, recompute_post, _ = helper_dir(
            out_posteriors, "path_posteriors", "", "posteriors"
        )
        out_resampled, recompute_resampled, _ = helper_dir(
            out_resampled, "path_resampled", "", "resampled"
        )
        out_volumes, recompute_volume, unique_volume_file = helper_dir(
            out_volumes, "path_volumes", "csv", ""
        )
        out_qc, recompute_qc, unique_qc_file = helper_dir(
            out_qc, "path_qc_scores", "csv", ""
        )

    # path_images is an image
    else:

        # input images
        assert os.path.isfile(path_images), (
            "file does not exist: %s \n"
            "please make sure the path and the extension are correct" % path_images
        )
        path_images = [path_images]

        # define helper to deal with outputs
        def helper_im(path, name, file_type, suffix):
            unique_file = False
            if path is not None:
                assert path[-4:] != ".txt", (
                    "%s can only be given as text file when path_images is." % name
                )
                if file_type == "csv":
                    if path[-4:] != ".csv":
                        print(
                            "%s provided without csv extension. Adding csv extension."
                            % name
                        )
                        path += ".csv"
                    recompute_files = [True]
                    unique_file = True
                else:
                    if (
                        (".nii.gz" not in path)
                        & (".nii" not in path)
                        & (".mgz" not in path)
                        & (".npz" not in path)
                    ):
                        file_name = os.path.basename(path_images[0]).replace(
                            ".nii", "_%s.nii" % suffix
                        )
                        file_name = file_name.replace(".mgz", "_%s.mgz" % suffix)
                        file_name = file_name.replace(".npz", "_%s.npz" % suffix)
                        path = os.path.join(path, file_name)
                    recompute_files = [not os.path.isfile(path)]
                utils.mkdir(os.path.dirname(path))
            else:
                recompute_files = [False]
            path = [path]
            return path, recompute_files, unique_file

        # use helper on all outputs
        out_seg, recompute_seg, _ = helper_im(
            out_seg, "path_segmentations", "", "synthseg"
        )
        out_posteriors, recompute_post, _ = helper_im(
            out_posteriors, "path_posteriors", "", "posteriors"
        )
        out_resampled, recompute_resampled, _ = helper_im(
            out_resampled, "path_resampled", "", "resampled"
        )
        out_volumes, recompute_volume, unique_volume_file = helper_im(
            out_volumes, "path_volumes", "csv", ""
        )
        out_qc, recompute_qc, unique_qc_file = helper_im(
            out_qc, "path_qc_scores", "csv", ""
        )

    recompute_list = [
        recompute | re_seg | re_post | re_res | re_vol | re_qc
        for (re_seg, re_post, re_res, re_vol, re_qc) in zip(
            recompute_seg,
            recompute_post,
            recompute_resampled,
            recompute_volume,
            recompute_qc,
        )
    ]

    return (
        path_images,
        out_seg,
        out_posteriors,
        out_resampled,
        out_volumes,
        unique_volume_file,
        out_qc,
        unique_qc_file,
        recompute_list,
    )


def preprocess(
    path_image,
    ct,
    target_res=1.0,
    n_levels=5,
    crop=None,
    min_pad=None,
    path_resample=None,
):

    # read image and corresponding info
    im, _, aff, n_dims, n_channels, h, im_res = utils.get_volume_info(path_image, True)
    if n_dims == 2 and 1 < n_channels < 4:
        raise Exception(
            "either the input is 2D with several channels, or is 3D with at most 3 slices. "
            "Either way, results are going to be poor..."
        )
    elif n_dims == 2 and 3 < n_channels < 11:
        print("warning: input with very few slices")
        n_dims = 3
    elif n_dims < 3:
        raise Exception("input should have 3 dimensions, had %s" % n_dims)
    elif n_dims == 4 and n_channels == 1:
        n_dims = 3
        im = im[..., 0]
    elif n_dims > 3:
        raise Exception("input should have 3 dimensions, had %s" % n_dims)
    elif n_channels > 1:
        print("WARNING: detected more than 1 channel, only keeping the first channel.")
        im = im[..., 0]

    # resample image if necessary
    target_res = np.squeeze(utils.reformat_to_n_channels_array(target_res, n_dims))
    if np.any((im_res > target_res + 0.05) | (im_res < target_res - 0.05)):
        im_res = target_res
        im, aff = edit_volumes.resample_volume(im, aff, im_res)
        if path_resample is not None:
            utils.save_volume(im, aff, h, path_resample)

    # align image
    im = edit_volumes.align_volume_to_ref(
        im, aff, aff_ref=np.eye(4), n_dims=n_dims, return_copy=False
    )
    shape = list(im.shape[:n_dims])

    # crop image if necessary
    if crop is not None:
        crop = utils.reformat_to_list(crop, length=n_dims, dtype="int")
        crop_shape = [
            utils.find_closest_number_divisible_by_m(s, 2**n_levels, "higher")
            for s in crop
        ]
        im, crop_idx = edit_volumes.crop_volume(
            im, cropping_shape=crop_shape, return_crop_idx=True
        )
    else:
        crop_idx = None

    # normalise image
    if ct:
        im = np.clip(im, 0, 80)
    im = edit_volumes.rescale_volume(
        im, new_min=0.0, new_max=1.0, min_percentile=0.5, max_percentile=99.5
    )

    # pad image
    input_shape = im.shape[:n_dims]
    pad_shape = [
        utils.find_closest_number_divisible_by_m(s, 2**n_levels, "higher")
        for s in input_shape
    ]
    min_pad = utils.reformat_to_list(min_pad, length=n_dims, dtype="int")
    min_pad = [
        utils.find_closest_number_divisible_by_m(s, 2**n_levels, "higher")
        for s in min_pad
    ]
    pad_shape = np.maximum(pad_shape, min_pad)
    im, pad_idx = edit_volumes.pad_volume(
        im, padding_shape=pad_shape, return_pad_idx=True
    )

    # add batch and channel axes
    im = utils.add_axis(im, axis=[0, -1])

    return im, aff, h, im_res, shape, pad_idx, crop_idx


def build_model(
    path_model_segmentation,
    path_model_parcellation,
    path_model_qc,
    input_shape_qc,
    labels_segmentation,
    labels_denoiser,
    labels_parcellation,
    labels_qc,
    sigma_smoothing,
    flip_indices,
    robust,
    do_parcellation,
    do_qc,
):

    # Ensure all model files exist by reassembling them if necessary
    if not ensure_model_file_exists(path_model_segmentation):
        raise FileNotFoundError(
            f"Could not find or reassemble model file: {path_model_segmentation}"
        )

    if do_parcellation and not ensure_model_file_exists(path_model_parcellation):
        raise FileNotFoundError(
            f"Could not find or reassemble parcellation model: {path_model_parcellation}"
        )

    if do_qc and not ensure_model_file_exists(path_model_qc):
        raise FileNotFoundError(
            f"Could not find or reassemble QC model: {path_model_qc}"
        )

    # Continue with the existing code...
    assert os.path.isfile(
        path_model_segmentation
    ), "The provided model path does not exist."

    # get labels
    n_labels_seg = len(labels_segmentation)

    if robust:
        n_groups = len(labels_denoiser)

        # build first UNet
        net = nrn_models.unet(
            input_shape=[None, None, None, 1],
            nb_labels=n_groups,
            nb_levels=5,
            nb_conv_per_level=2,
            conv_size=3,
            nb_features=24,
            feat_mult=2,
            activation="elu",
            batch_norm=-1,
            name="unet",
        )

        # transition between the two networks: one_hot -> argmax -> one_hot (it simulates how the network was trained)
        last_tensor = net.output
        last_tensor = KL.Lambda(lambda x: tf.argmax(x, axis=-1))(last_tensor)
        last_tensor = KL.Lambda(
            lambda x: tf.one_hot(tf.cast(x, "int32"), depth=n_groups, axis=-1)
        )(last_tensor)
        net = Model(inputs=net.inputs, outputs=last_tensor)

        # build denoiser
        net = nrn_models.unet(
            input_model=net,
            input_shape=[None, None, None, 1],
            nb_labels=n_groups,
            nb_levels=5,
            nb_conv_per_level=2,
            conv_size=5,
            nb_features=16,
            feat_mult=2,
            activation="elu",
            batch_norm=-1,
            skip_n_concatenations=2,
            name="l2l",
        )

        # transition between the two networks: one_hot -> argmax -> one_hot, and concatenate input image and labels
        input_image = net.inputs[0]
        last_tensor = net.output
        last_tensor = KL.Lambda(lambda x: tf.argmax(x, axis=-1))(last_tensor)
        last_tensor = KL.Lambda(
            lambda x: tf.one_hot(tf.cast(x, "int32"), depth=n_groups, axis=-1)
        )(last_tensor)
        if n_groups <= 2:
            last_tensor = KL.Lambda(lambda x: x[..., 1:])(last_tensor)
        last_tensor = KL.Lambda(lambda x: tf.cast(tf.concat(x, axis=-1), "float32"))(
            [input_image, last_tensor]
        )
        net = Model(inputs=net.inputs, outputs=last_tensor)

        # build 2nd network
        net = nrn_models.unet(
            input_model=net,
            input_shape=[None, None, None, 2],
            nb_labels=n_labels_seg,
            nb_levels=5,
            nb_conv_per_level=2,
            conv_size=3,
            nb_features=24,
            feat_mult=2,
            activation="elu",
            batch_norm=-1,
            name="unet2",
        )
        net.load_weights(path_model_segmentation, by_name=True)
        name_segm_prediction_layer = "unet2_prediction"

    else:

        # build UNet
        net = nrn_models.unet(
            input_shape=[None, None, None, 1],
            nb_labels=n_labels_seg,
            nb_levels=5,
            nb_conv_per_level=2,
            conv_size=3,
            nb_features=24,
            feat_mult=2,
            activation="elu",
            batch_norm=-1,
            name="unet",
        )
        net.load_weights(path_model_segmentation, by_name=True)
        input_image = net.inputs[0]
        name_segm_prediction_layer = "unet_prediction"

        # smooth posteriors if specified
        if sigma_smoothing > 0:
            last_tensor = net.output
            if hasattr(last_tensor, "get_shape"):
                last_tensor._keras_shape = tuple(last_tensor.get_shape().as_list())
            last_tensor = layers.GaussianBlur(sigma=sigma_smoothing)(last_tensor)
            net = Model(inputs=net.inputs, outputs=last_tensor)

        if flip_indices is not None:

            # segment flipped image
            seg = net.output
            image_flipped = layers.RandomFlip(axis=0, prob=1)(input_image)
            last_tensor = net(image_flipped)

            # flip back and re-order channels
            last_tensor = layers.RandomFlip(axis=0, prob=1)(last_tensor)
            last_tensor = KL.Lambda(
                lambda x: tf.split(x, [1] * n_labels_seg, axis=-1), name="split"
            )(last_tensor)
            reordered_channels = [
                last_tensor[flip_indices[i]] for i in range(n_labels_seg)
            ]
            last_tensor = KL.Lambda(lambda x: tf.concat(x, -1), name="concat")(
                reordered_channels
            )

            # average two segmentations and build model
            name_segm_prediction_layer = "average_lr"
            last_tensor = KL.Lambda(
                lambda x: 0.5 * (x[0] + x[1]), name=name_segm_prediction_layer
            )([seg, last_tensor])
            net = Model(inputs=net.inputs, outputs=last_tensor)

    # add aparc segmenter if needed
    if do_parcellation:
        n_labels_parcellation = len(labels_parcellation)

        # build input for S3: only takes one map for cortical segmentation (no image), 1 = cortex, 0 = other
        last_tensor = net.output
        last_tensor = KL.Lambda(lambda x: tf.cast(tf.argmax(x, axis=-1), "int32"))(
            last_tensor
        )
        last_tensor = layers.ConvertLabels(
            np.arange(n_labels_seg), labels_segmentation
        )(last_tensor)
        parcellation_masking_values = np.array(
            [1 if ((ll == 3) | (ll == 42)) else 0 for ll in labels_segmentation]
        )
        last_tensor = layers.ConvertLabels(
            labels_segmentation, parcellation_masking_values
        )(last_tensor)
        last_tensor = KL.Lambda(
            lambda x: tf.one_hot(tf.cast(x, "int32"), depth=2, axis=-1)
        )(last_tensor)
        last_tensor = KL.Lambda(lambda x: tf.cast(tf.concat(x, axis=-1), "float32"))(
            [input_image, last_tensor]
        )
        net = Model(inputs=net.inputs, outputs=last_tensor)

        # build UNet
        net = nrn_models.unet(
            input_model=net,
            input_shape=[None, None, None, 3],
            nb_labels=n_labels_parcellation,
            nb_levels=5,
            nb_conv_per_level=2,
            conv_size=3,
            nb_features=24,
            feat_mult=2,
            activation="elu",
            batch_norm=-1,
            name="unet_parc",
        )
        net.load_weights(path_model_parcellation, by_name=True)

        # smooth predictions
        last_tensor = net.output
        if hasattr(last_tensor, "get_shape"):
            last_tensor._keras_shape = tuple(last_tensor.get_shape().as_list())
        last_tensor = layers.GaussianBlur(sigma=0.5)(last_tensor)
        net = Model(
            inputs=net.inputs,
            outputs=[net.get_layer(name_segm_prediction_layer).output, last_tensor],
        )

    # add CNN regressor for automated QC if needed
    if do_qc:
        n_labels_qc = len(np.unique(labels_qc))

        # transition between the two networks: one_hot -> argmax -> qc_labels -> one_hot
        shape_prediction = KL.Input([3], dtype="int32")
        if do_parcellation:
            last_tensor = KL.Lambda(
                lambda x: tf.cast(tf.argmax(x[0], axis=-1), "int32")
            )(net.outputs)
        else:
            last_tensor = KL.Lambda(lambda x: tf.cast(tf.argmax(x, axis=-1), "int32"))(
                net.output
            )
        last_tensor = MakeShape(input_shape_qc)([last_tensor, shape_prediction])
        last_tensor = layers.ConvertLabels(
            np.arange(n_labels_seg), labels_segmentation
        )(last_tensor)
        last_tensor = layers.ConvertLabels(labels_segmentation, labels_qc)(last_tensor)
        last_tensor = KL.Lambda(
            lambda x: tf.one_hot(tf.cast(x, "int32"), depth=n_labels_qc, axis=-1)
        )(last_tensor)
        net = Model(inputs=[*net.inputs, shape_prediction], outputs=last_tensor)

        # build QC regressor network
        net = nrn_models.conv_enc(
            input_model=net,
            input_shape=[None, None, None, 1],
            nb_levels=4,
            nb_conv_per_level=2,
            conv_size=5,
            nb_features=24,
            feat_mult=2,
            activation="relu",
            batch_norm=-1,
            use_residuals=True,
            name="qc",
        )
        last_tensor = net.outputs[0]
        conv_kwargs = {
            "padding": "same",
            "activation": "relu",
            "data_format": "channels_last",
        }
        last_tensor = KL.MaxPool3D(
            pool_size=(2, 2, 2), name="qc_maxpool_3", padding="same"
        )(last_tensor)
        last_tensor = KL.Conv3D(
            n_labels_qc, kernel_size=5, **conv_kwargs, name="qc_final_conv_0"
        )(last_tensor)
        last_tensor = KL.Conv3D(
            n_labels_qc, kernel_size=5, **conv_kwargs, name="qc_final_conv_1"
        )(last_tensor)
        last_tensor = KL.Lambda(
            lambda x: tf.reduce_mean(x, axis=[1, 2, 3]), name="qc_final_pred"
        )(last_tensor)

        # build model
        if do_parcellation:
            outputs = [
                net.get_layer(name_segm_prediction_layer).output,
                net.get_layer("unet_parc_prediction").output,
                last_tensor,
            ]
        else:
            outputs = [net.get_layer(name_segm_prediction_layer).output, last_tensor]
        net = Model(inputs=net.inputs, outputs=outputs)
        net.load_weights(path_model_qc, by_name=True)

    return net


def postprocess(
    post_patch_seg,
    post_patch_parc,
    shape,
    pad_idx,
    crop_idx,
    labels_segmentation,
    labels_parcellation,
    aff,
    im_res,
    fast,
    topology_classes,
    v1,
):

    # get posteriors
    post_patch_seg = np.squeeze(post_patch_seg)
    if fast | (topology_classes is None):
        post_patch_seg = edit_volumes.crop_volume_with_idx(
            post_patch_seg, pad_idx, n_dims=3, return_copy=False
        )

    # keep biggest connected component
    tmp_post_patch_seg = post_patch_seg[..., 1:]
    post_patch_seg_mask = np.sum(tmp_post_patch_seg, axis=-1) > 0.25
    post_patch_seg_mask = edit_volumes.get_largest_connected_component(
        post_patch_seg_mask
    )
    post_patch_seg_mask = np.stack(
        [post_patch_seg_mask] * tmp_post_patch_seg.shape[-1], axis=-1
    )
    tmp_post_patch_seg = edit_volumes.mask_volume(
        tmp_post_patch_seg, mask=post_patch_seg_mask, return_copy=False
    )
    post_patch_seg[..., 1:] = tmp_post_patch_seg

    # reset posteriors to zero outside the largest connected component of each topological class
    if (not fast) & (topology_classes is not None):
        post_patch_seg_mask = post_patch_seg > 0.25
        for topology_class in np.unique(topology_classes)[1:]:
            tmp_topology_indices = np.where(topology_classes == topology_class)[0]
            tmp_mask = np.any(post_patch_seg_mask[..., tmp_topology_indices], axis=-1)
            tmp_mask = edit_volumes.get_largest_connected_component(tmp_mask)
            for idx in tmp_topology_indices:
                post_patch_seg[..., idx] *= tmp_mask
        post_patch_seg = edit_volumes.crop_volume_with_idx(
            post_patch_seg, pad_idx, n_dims=3, return_copy=False
        )
    else:
        post_patch_seg_mask = post_patch_seg > 0.2
        post_patch_seg[..., 1:] *= post_patch_seg_mask[..., 1:]

    # get hard segmentation
    post_patch_seg /= np.sum(post_patch_seg, axis=-1)[..., np.newaxis]
    seg_patch = labels_segmentation[post_patch_seg.argmax(-1).astype("int32")].astype(
        "int32"
    )

    # postprocess parcellation
    if post_patch_parc is not None:
        post_patch_parc = np.squeeze(post_patch_parc)
        post_patch_parc = edit_volumes.crop_volume_with_idx(
            post_patch_parc, pad_idx, n_dims=3, return_copy=False
        )
        mask = (seg_patch == 3) | (seg_patch == 42)
        post_patch_parc[..., 0] = np.ones_like(post_patch_parc[..., 0])
        post_patch_parc[..., 0] = edit_volumes.mask_volume(
            post_patch_parc[..., 0], mask=mask < 0.1, return_copy=False
        )
        post_patch_parc /= np.sum(post_patch_parc, axis=-1)[..., np.newaxis]
        parc_patch = labels_parcellation[
            post_patch_parc.argmax(-1).astype("int32")
        ].astype("int32")
        seg_patch[mask] = parc_patch[mask]

    # paste patches back to matrix of original image size
    if crop_idx is not None:
        # we need to go through this because of the posteriors of the background, otherwise pad_volume would work
        seg = np.zeros(shape=shape, dtype="int32")
        posteriors = np.zeros(shape=[*shape, labels_segmentation.shape[0]])
        posteriors[..., 0] = np.ones(shape)  # place background around patch
        seg[
            crop_idx[0] : crop_idx[3],
            crop_idx[1] : crop_idx[4],
            crop_idx[2] : crop_idx[5],
        ] = seg_patch
        posteriors[
            crop_idx[0] : crop_idx[3],
            crop_idx[1] : crop_idx[4],
            crop_idx[2] : crop_idx[5],
            :,
        ] = post_patch_seg
    else:
        seg = seg_patch
        posteriors = post_patch_seg

    # align prediction back to first orientation
    seg = edit_volumes.align_volume_to_ref(
        seg, aff=np.eye(4), aff_ref=aff, n_dims=3, return_copy=False
    )
    posteriors = edit_volumes.align_volume_to_ref(
        posteriors, np.eye(4), aff_ref=aff, n_dims=3, return_copy=False
    )

    # compute volumes
    volumes = np.sum(
        posteriors[..., 1:], axis=tuple(range(0, len(posteriors.shape) - 1))
    )
    total_volume_cortex_left = np.sum(
        volumes[np.where(labels_segmentation == 3)[0] - 1]
    )
    total_volume_cortex_right = np.sum(
        volumes[np.where(labels_segmentation == 42)[0] - 1]
    )
    if not v1:
        volumes = np.concatenate([np.array([np.sum(volumes)]), volumes])
    if post_patch_parc is not None:
        volumes_parc = np.sum(
            post_patch_parc[..., 1:], axis=tuple(range(0, len(posteriors.shape) - 1))
        )
        volumes_parc_left = volumes_parc[: int(len(volumes_parc) / 2)]
        volumes_parc_right = volumes_parc[int(len(volumes_parc) / 2) :]
        volumes_parc_left = (
            volumes_parc_left / np.sum(volumes_parc_left) * total_volume_cortex_left
        )
        volumes_parc_right = (
            volumes_parc_right / np.sum(volumes_parc_right) * total_volume_cortex_right
        )
        volumes = np.concatenate([volumes, volumes_parc_left, volumes_parc_right])
    volumes = np.around(volumes * np.prod(im_res), 3)

    return seg, posteriors, volumes


class MakeShape(KL.Layer):
    """Expects one-hot encoding of the two input label maps."""

    def __init__(self, target_shape, **kwargs):
        self.n_dims = None
        self.target_shape = target_shape
        self.cropping_shape = None
        super(MakeShape, self).__init__(**kwargs)

    def get_config(self):
        config = super().get_config()
        config["target_shape"] = self.target_shape
        return config

    def build(self, input_shape):
        self.n_dims = input_shape[1][1]
        self.cropping_shape = np.array(
            utils.reformat_to_list(self.target_shape, length=self.n_dims)
        )
        self.built = True
        super(MakeShape, self).build(input_shape)

    def call(self, inputs, **kwargs):
        return tf.map_fn(self._single_process, inputs, dtype=tf.int32)

    def _single_process(self, inputs):

        x = inputs[0]
        shape = inputs[1]

        # find cropping indices
        mask = tf.logical_and(tf.not_equal(x, 0), tf.not_equal(x, 24))
        indices = tf.cast(tf.where(mask), "int32")

        min_idx = K.switch(
            tf.equal(tf.shape(indices)[0], 0),
            tf.zeros(self.n_dims, dtype="int32"),
            tf.maximum(tf.reduce_min(indices, axis=0), 0),
        )
        max_idx = K.switch(
            tf.equal(tf.shape(indices)[0], 0),
            tf.minimum(shape, self.cropping_shape),
            tf.minimum(tf.reduce_max(indices, axis=0) + 1, shape),
        )

        # expand/retract (depending on the desired shape) the cropping region around the centre
        intermediate_vol_shape = max_idx - min_idx
        min_idx = min_idx - tf.cast(
            tf.math.ceil((self.cropping_shape - intermediate_vol_shape) / 2), "int32"
        )
        max_idx = max_idx + tf.cast(
            tf.math.floor((self.cropping_shape - intermediate_vol_shape) / 2), "int32"
        )
        tmp_min_idx = tf.maximum(min_idx, 0)
        tmp_max_idx = tf.minimum(max_idx, shape)
        x = tf.slice(
            x, begin=tmp_min_idx, size=tf.minimum(tmp_max_idx - tmp_min_idx, shape)
        )

        # pad if necessary
        min_padding = tf.abs(tf.minimum(min_idx, 0))
        max_padding = tf.maximum(max_idx - shape, 0)
        x = K.switch(
            tf.reduce_any(
                tf.logical_or(tf.greater(min_padding, 0), tf.greater(max_padding, 0))
            ),
            tf.pad(x, tf.stack([min_padding, max_padding], axis=1)),
            x,
        )

        return x



"""

This function trains a regressor network to predict Dice scores between segmentations (typically obtained with an
automated algorithm), and their ground truth (to which wwe typically do not have access at test time).

If you use this code, please cite one of the SynthSeg papers:
https://github.com/BBillot/SynthSeg/blob/master/bibtex.bib

Copyright 2020 Benjamin Billot

Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software distributed under the License is
distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
implied. See the License for the specific language governing permissions and limitations under the
License.
"""

# python imports
import os
import keras
import numpy as np
import tensorflow as tf
from keras import models
import keras.layers as KL
import keras.backend as K
import keras.callbacks as KC
from keras.optimizers import Adam
from inspect import getmembers, isclass

# project imports
from lamareg.SynthSeg import metrics_model as metrics

# third-party imports
from lamareg.ext.lab2im import utils
from lamareg.ext.lab2im import layers as l2i_layers
from lamareg.ext.neuron import utils as nrn_utils
from lamareg.ext.neuron import layers as nrn_layers
from lamareg.ext.neuron import models as nrn_models


def training(
    list_paths_input_labels,
    list_paths_target_labels,
    model_dir,
    labels_list,
    labels_list_to_convert=None,
    subjects_prob=None,
    batchsize=1,
    output_shape=None,
    scaling_bounds=0.2,
    rotation_bounds=15,
    shearing_bounds=0.012,
    translation_bounds=False,
    nonlin_std=4.0,
    nonlin_scale=0.04,
    n_levels=5,
    nb_conv_per_level=3,
    conv_size=5,
    unet_feat_count=24,
    feat_multiplier=2,
    activation="relu",
    lr=1e-4,
    epochs=300,
    steps_per_epoch=1000,
    checkpoint=None,
):
    """
    This function trains a regressor network to predict Dice scores between segmentations (typically obtained with an
    automated algorithm), and their ground truth (to which we typically do not have access at test time).

    # IMPORTANT !!!
    # Each time we provide a parameter with separate values for each axis (e.g. with a numpy array or a sequence),
    # these values refer to the RAS axes.

    :param list_paths_input_labels: list of all the paths of the input label maps. These typically correspond to the
    outputs given by a segmentation algorithm, for which the network will try to predict quality scores.
    :param list_paths_target_labels: list of all the paths of the ground truths for the input label maps. These are used
    to compute Dice scores with their corresponding input label maps, that will serve as target for the regression
    network. As such, target label maps must have the same label values as the input label maps.
    :param model_dir: path of a directory where the models will be saved during training.
    :param labels_list: list of all the label values present in the input/GT label maps.

    #---------------------------------------------- Generation parameters ----------------------------------------------
    # label maps parameters
    :param labels_list_to_convert: (optional) instead of regressing Dice scores for all the labels present in the
    input/GT label maps, one can "regroup" them into "groups". This must be a sequence or a 1d numpy array of the same
    length as labels_list with the corresponding group. By default (None), regression scores are computed for all labels
    :param subjects_prob: (optional) relative order of importance (doesn't have to be probabilistic), with which to pick
    the provided label maps at each minibatch. Can be a sequence, a 1D numpy array, or the path to such an array, and it
    must be as long as path_label_maps. By default, all label maps are chosen with the same importance.

    # output-related parameters
    :param batchsize: (optional) number of images to generate per mini-batch. Default is 1.
    :param output_shape: (optional) desired shape of the output image, obtained by randomly cropping the generated image
    Can be an integer (same size in all dimensions), a sequence, a 1d numpy array, or the path to a 1d numpy array.
    Default is None, where no cropping is performed.

    # spatial deformation parameters
    :param scaling_bounds: (optional) if apply_linear_trans is True, the scaling factor for each dimension is
    sampled from a uniform distribution of predefined bounds. Can either be:
    1) a number, in which case the scaling factor is independently sampled from the uniform distribution of bounds
    (1-scaling_bounds, 1+scaling_bounds) for each dimension.
    2) the path to a numpy array of shape (2, n_dims), in which case the scaling factor in dimension i is sampled from
    the uniform distribution of bounds (scaling_bounds[0, i], scaling_bounds[1, i]) for the i-th dimension.
    3) False, in which case scaling is completely turned off.
    Default is scaling_bounds = 0.2 (case 1)
    :param rotation_bounds: (optional) same as scaling bounds but for the rotation angle, except that for case 1 the
    bounds are centred on 0 rather than 1, i.e. (0+rotation_bounds[i], 0-rotation_bounds[i]).
    Default is rotation_bounds = 15.
    :param shearing_bounds: (optional) same as scaling bounds. Default is shearing_bounds = 0.012.
    :param translation_bounds: (optional) same as scaling bounds. Default is translation_bounds = False, but we
    encourage using it when cropping is deactivated (i.e. when output_shape=None).
    :param nonlin_std: (optional) Standard deviation of the normal distribution from which we sample the first
    tensor for synthesising the deformation field. Set to 0 to completely deactivate elastic deformation.
    :param nonlin_scale: (optional) Ratio between the size of the input label maps and the size of the sampled
    tensor for synthesising the elastic deformation field.

    # ------------------------------------------ UNet architecture parameters ------------------------------------------
    :param n_levels: (optional) number of level for the Unet. Default is 5.
    :param nb_conv_per_level: (optional) number of convolutional layers per level. Default is 2.
    :param conv_size: (optional) size of the convolution kernels. Default is 2.
    :param unet_feat_count: (optional) number of feature for the first layer of the UNet. Default is 24.
    :param feat_multiplier: (optional) multiply the number of feature by this number at each new level. Default is 2.
    :param activation: (optional) activation function. Can be 'elu', 'relu'.

    # ----------------------------------------------- Training parameters ----------------------------------------------
    :param lr: (optional) learning rate for the training. Default is 1e-4
    :param epochs: (optional) number of training epochs. Default is 300.
    :param steps_per_epoch: (optional) number of steps per epoch. Default is 10000. Since no online validation is
    possible, this is equivalent to the frequency at which the models are saved.
    :param checkpoint: (optional) path of an already saved model to load before starting the training.
    """

    # prepare data files
    labels_list, _ = utils.get_list_labels(label_list=labels_list)
    if labels_list_to_convert is not None:
        labels_list_to_convert, _ = utils.get_list_labels(
            label_list=labels_list_to_convert
        )
    n_labels = len(np.unique(labels_list))

    # create augmentation model
    labels_shape, _, n_dims, _, _, _ = utils.get_volume_info(
        list_paths_target_labels[0], aff_ref=np.eye(4)
    )
    augmentation_model = build_augmentation_model(
        labels_shape,
        labels_list,
        labels_list_to_convert=labels_list_to_convert,
        output_shape=output_shape,
        output_div_by_n=2**n_levels,
        scaling_bounds=scaling_bounds,
        rotation_bounds=rotation_bounds,
        shearing_bounds=shearing_bounds,
        translation_bounds=translation_bounds,
        nonlin_std=nonlin_std,
        nonlin_scale=nonlin_scale,
    )

    # prepare QC model
    regression_model = build_qc_model(
        input_model=augmentation_model,
        n_labels=n_labels,
        n_levels=n_levels,
        nb_conv_per_level=nb_conv_per_level,
        conv_size=conv_size,
        unet_feat_count=unet_feat_count,
        feat_multiplier=feat_multiplier,
        activation=activation,
    )
    qc_model = build_qc_loss(regression_model)

    # input generator
    model_inputs = build_model_inputs(
        path_input_label_maps=list_paths_input_labels,
        path_target_label_maps=list_paths_target_labels,
        batchsize=batchsize,
        subjects_prob=subjects_prob,
    )
    input_generator = utils.build_training_generator(model_inputs, batchsize)

    train_model(
        qc_model,
        input_generator,
        lr,
        epochs,
        steps_per_epoch,
        model_dir,
        "qc",
        checkpoint,
    )


def build_augmentation_model(
    labels_shape,
    labels_list,
    labels_list_to_convert=None,
    output_shape=None,
    output_div_by_n=None,
    scaling_bounds=0.15,
    rotation_bounds=15,
    shearing_bounds=0.012,
    translation_bounds=False,
    nonlin_std=3.0,
    nonlin_scale=0.0625,
):

    # reformat resolutions and get shapes
    labels_shape = utils.reformat_to_list(labels_shape)
    n_dims, _ = utils.get_dims(labels_shape)

    # get shapes
    output_shape = get_shapes(labels_shape, output_shape, output_div_by_n, n_dims)

    # define model inputs
    net_input = KL.Input(
        shape=labels_shape + [1], name="noisy_labels_input", dtype="int32"
    )
    target_input = KL.Input(
        shape=labels_shape + [1], name="target_input", dtype="int32"
    )

    # convert labels if necessary
    if labels_list_to_convert is not None:
        noisy_labels = l2i_layers.ConvertLabels(
            labels_list_to_convert, labels_list, name="convert_noisy"
        )(net_input)
        target = l2i_layers.ConvertLabels(
            labels_list_to_convert, labels_list, name="convert_target"
        )(target_input)
    else:
        noisy_labels = net_input
        target = target_input

    # deform labels
    noisy_labels, target = l2i_layers.RandomSpatialDeformation(
        scaling_bounds=scaling_bounds,
        rotation_bounds=rotation_bounds,
        shearing_bounds=shearing_bounds,
        translation_bounds=translation_bounds,
        nonlin_std=nonlin_std,
        nonlin_scale=nonlin_scale,
        inter_method="nearest",
    )([noisy_labels, target])

    # mask image, compute Dice score with full GT, and crop noisy labels
    noisy_labels, scores = SimulatePartialFOV(
        crop_shape=output_shape[0],
        labels_list=np.unique(labels_list),
        min_fov_shape=70,
        prob_mask=0.3,
        name="partial_fov",
    )([noisy_labels, target])

    # dummy layers
    scores = KL.Lambda(lambda x: x, name="dice_gt")(scores)
    noisy_labels = KL.Lambda(lambda x: x[0], name="labels_out")([noisy_labels, scores])

    # build model and return
    brain_model = models.Model(inputs=[net_input, target_input], outputs=noisy_labels)
    return brain_model


def build_qc_model(
    input_model,
    n_labels,
    n_levels,
    nb_conv_per_level,
    conv_size,
    unet_feat_count,
    feat_multiplier,
    activation,
):

    # get prediction
    last_tensor = input_model.outputs[0]
    input_shape = last_tensor.get_shape().as_list()[1:]
    assert (
        input_shape[-1] == n_labels
    ), "mismatch between number of predicted labels, and segmentation labels"

    # build model
    model = nrn_models.conv_enc(
        input_model=input_model,
        input_shape=input_shape,
        nb_levels=n_levels,
        nb_conv_per_level=nb_conv_per_level,
        conv_size=conv_size,
        nb_features=unet_feat_count,
        feat_mult=feat_multiplier,
        activation=activation,
        batch_norm=-1,
        use_residuals=True,
        name="qc",
    )
    last = model.outputs[0]

    conv_kwargs = {
        "padding": "same",
        "activation": "relu",
        "data_format": "channels_last",
    }
    last = KL.MaxPool3D(
        pool_size=(2, 2, 2), name="qc_maxpool_%s" % (n_levels - 1), padding="same"
    )(last)
    last = KL.Conv3D(n_labels, kernel_size=5, **conv_kwargs, name="qc_final_conv_0")(
        last
    )
    last = KL.Conv3D(n_labels, kernel_size=5, **conv_kwargs, name="qc_final_conv_1")(
        last
    )
    last = KL.Lambda(lambda x: tf.reduce_mean(x, axis=[1, 2, 3]), name="qc_final_pred")(
        last
    )

    return models.Model(input_model.inputs, last)


def build_qc_loss(input_model):

    # get Dice scores
    dice_gt = input_model.get_layer("dice_gt").output
    dice_pred = input_model.outputs[0]

    # get loss
    loss = KL.Lambda(
        lambda x: K.sum(K.mean(K.square(x[0] - x[1]), axis=0)), name="qc_loss"
    )([dice_gt, dice_pred])
    loss._keras_shape = tuple(loss.get_shape().as_list())

    return models.Model(inputs=input_model.inputs, outputs=loss)


def build_model_inputs(
    path_input_label_maps, path_target_label_maps, batchsize=1, subjects_prob=None
):

    # make sure subjects_prob sums to 1
    subjects_prob = utils.load_array_if_path(subjects_prob)
    if subjects_prob is not None:
        subjects_prob /= np.sum(subjects_prob)

    # Generate!
    while True:

        # randomly pick as many images as batchsize
        indices = np.random.choice(
            np.arange(len(path_input_label_maps)), size=batchsize, p=subjects_prob
        )

        # initialise input lists
        list_input_label_maps = list()
        list_target_label_maps = list()

        for idx in indices:

            # load input
            input_net = utils.load_volume(
                path_input_label_maps[idx], dtype="int", aff_ref=np.eye(4)
            )
            list_input_label_maps.append(utils.add_axis(input_net, axis=[0, -1]))

            # load target
            target = utils.load_volume(
                path_target_label_maps[idx], dtype="int", aff_ref=np.eye(4)
            )
            list_target_label_maps.append(utils.add_axis(target, axis=[0, -1]))

        # build list of training pairs
        list_training_pairs = [list_input_label_maps, list_target_label_maps]
        if batchsize > 1:  # concatenate individual input types if batchsize > 1
            list_training_pairs = [
                np.concatenate(item, 0) for item in list_training_pairs
            ]
        else:
            list_training_pairs = [item[0] for item in list_training_pairs]

        yield list_training_pairs


def get_shapes(labels_shape, cropping_shape, output_div_by_n, n_dims):

    # cropping shape specified, make sure it's okay
    if cropping_shape is not None:
        cropping_shape = utils.reformat_to_list(
            cropping_shape, length=n_dims, dtype="int"
        )

        # make sure that cropping shape is smaller or equal to label shape
        cropping_shape = [
            min(labels_shape[i], cropping_shape[i]) for i in range(n_dims)
        ]

        # make sure cropping shape is divisible by output_div_by_n
        if output_div_by_n is not None:
            tmp_shape = [
                utils.find_closest_number_divisible_by_m(s, output_div_by_n)
                for s in cropping_shape
            ]
            if cropping_shape != tmp_shape:
                print(
                    "output shape {0} not divisible by {1}, changed to {2}".format(
                        cropping_shape, output_div_by_n, tmp_shape
                    )
                )
                cropping_shape = tmp_shape

    # no cropping shape specified, so no cropping unless label_shape is not divisible by output_div_by_n
    else:

        # make sure labels shape is divisible by output_div_by_n
        if output_div_by_n is not None:
            cropping_shape = [
                utils.find_closest_number_divisible_by_m(s, output_div_by_n)
                for s in labels_shape
            ]

        # if no need to be divisible by n, simply take cropping_shape as image_shape, and build output_shape
        else:
            cropping_shape = labels_shape

    return cropping_shape


def train_model(
    model,
    generator,
    learning_rate,
    n_epochs,
    n_steps,
    model_dir,
    metric_type,
    path_checkpoint=None,
    reinitialise_momentum=False,
):

    # prepare model and log folders
    utils.mkdir(model_dir)
    log_dir = os.path.join(model_dir, "logs")
    utils.mkdir(log_dir)

    # model saving callback
    save_file_name = os.path.join(model_dir, "qc_{epoch:03d}.h5")
    callbacks = [
        KC.ModelCheckpoint(save_file_name, verbose=1),
        KC.TensorBoard(
            log_dir=log_dir, histogram_freq=0, write_graph=True, write_images=False
        ),
    ]

    compile_model = True
    init_epoch = 0
    if (path_checkpoint is not None) & (not reinitialise_momentum):
        init_epoch = int(os.path.basename(path_checkpoint).split(metric_type)[1][1:-3])
        custom_l2i = {
            key: value
            for (key, value) in getmembers(l2i_layers, isclass)
            if key != "Layer"
        }
        custom_nrn = {
            key: value
            for (key, value) in getmembers(nrn_layers, isclass)
            if key != "Layer"
        }
        custom_objects = {
            **custom_l2i,
            **custom_nrn,
            "tf": tf,
            "keras": keras,
            "loss": metrics.IdentityLoss().loss,
        }
        model = models.load_model(path_checkpoint, custom_objects=custom_objects)
        compile_model = False
    elif path_checkpoint is not None:
        model.load_weights(path_checkpoint, by_name=True)

    # compile
    if compile_model:
        model.compile(
            optimizer=Adam(lr=learning_rate), loss=metrics.IdentityLoss().loss
        )

    # fit
    model.fit_generator(
        generator,
        epochs=n_epochs,
        steps_per_epoch=n_steps,
        callbacks=callbacks,
        initial_epoch=init_epoch,
    )


class SimulatePartialFOV(KL.Layer):
    """Expects hard segmentations for the two input label maps, input first, gt second."""

    def __init__(self, crop_shape, labels_list, min_fov_shape, prob_mask, **kwargs):

        # cropping arguments
        self.crop_shape = crop_shape
        self.crop_max_val = None

        # asking arguments
        self.min_fov_shape = min_fov_shape
        self.prob_mask = prob_mask

        # shape arguments
        self.inshape = None
        self.n_dims = None
        self.labels_list = labels_list
        self.n_labels = len(labels_list)
        self.lut = None
        self.meshgrid = None

        super(SimulatePartialFOV, self).__init__(**kwargs)

    def get_config(self):
        config = super().get_config()
        config["labels_list"] = self.labels_list
        config["crop_shape"] = self.crop_shape
        config["min_fov_shape"] = self.min_fov_shape
        config["prob_mask"] = self.prob_mask
        return config

    def build(self, input_shape):

        # check shapes
        assert (
            len(input_shape) == 2
        ), "SimulatePartialFOV expects 2 inputs: labels to deform and GT (for Dice scores)."
        assert (
            input_shape[0] == input_shape[1]
        ), "the two inputs must have the same shape."
        self.n_dims = len(input_shape[0]) - 2
        self.inshape = input_shape[0][1 : self.n_dims + 1]

        self.crop_max_val = self.inshape[0] - self.crop_shape
        self.meshgrid = nrn_utils.volshape_to_ndgrid(self.inshape)
        self.lut = tf.convert_to_tensor(
            utils.get_mapping_lut(self.labels_list), dtype="int32"
        )

        self.built = True
        super(SimulatePartialFOV, self).build(input_shape)

    def call(self, inputs, **kwargs):

        # get inputs
        x = inputs[0][..., 0]
        y = inputs[1][..., 0]

        # sample cropping indices
        batchsize = tf.split(tf.shape(x), [1, -1])[0]
        sample_shape = tf.concat(
            [batchsize, self.n_dims * tf.ones([1], dtype="int32")], 0
        )
        if self.crop_max_val > 0:
            crop_idx_inf = tf.random.uniform(
                shape=sample_shape, minval=0, maxval=self.crop_max_val, dtype="int32"
            )
            crop_idx_sup = crop_idx_inf + self.crop_shape
        else:
            crop_idx_inf = crop_idx_sup = None

        # sample masking indices
        fov_shape = tf.random.uniform(
            sample_shape,
            minval=self.min_fov_shape,
            maxval=self.inshape[0],
            dtype="int32",
        )
        mask_idx_inf = tf.random.uniform(
            shape=sample_shape, minval=0, maxval=1, dtype="float32"
        )
        mask_idx_inf_tmp = tf.cast(
            mask_idx_inf * tf.cast(self.inshape[0] - fov_shape, "float32"), "int32"
        )
        mask_idx_sup_tmp = mask_idx_inf_tmp + fov_shape
        if self.crop_max_val > 0:
            mask_idx_inf = tf.maximum(mask_idx_inf_tmp, crop_idx_inf)
            mask_idx_sup = tf.minimum(mask_idx_sup_tmp, crop_idx_sup)
        else:
            mask_idx_inf = mask_idx_inf_tmp
            mask_idx_sup = mask_idx_sup_tmp

        # mask input labels
        mask = tf.map_fn(
            self._single_build_mask, [x, mask_idx_inf, mask_idx_sup], tf.int32
        )
        x = K.switch(
            tf.squeeze(K.greater(tf.random.uniform([1], 0, 1), 1 - self.prob_mask)),
            x * mask,
            x,
        )

        # compute dice score for each label value
        x = tf.one_hot(tf.gather(self.lut, x), depth=self.n_labels, axis=-1)
        y = tf.one_hot(tf.gather(self.lut, y), depth=self.n_labels, axis=-1)
        top = tf.math.reduce_sum(2 * x * y, axis=list(range(1, self.n_dims + 1)))
        bottom = x + y + tf.keras.backend.epsilon()
        bottom = tf.math.reduce_sum(bottom, axis=list(range(1, self.n_dims + 1)))
        dice_score = top / bottom

        # crop input labels
        if self.crop_max_val > 0:
            x_cropped = tf.map_fn(
                self._single_slice, [x, crop_idx_inf], dtype=tf.float32
            )
        else:
            x_cropped = x

        return [x_cropped, dice_score]

    def _single_build_mask(self, inputs):
        vol = inputs[0]
        mask_idx_inf = inputs[1]
        mask_idx_sup = inputs[2]
        mask = tf.ones(vol.shape, dtype="bool")
        for i in range(self.n_dims):
            tmp_mask_inf = tf.less(self.meshgrid[i], mask_idx_inf[i])
            tmp_mask_sup = tf.greater(self.meshgrid[i], mask_idx_sup[i])
            mask = tf.logical_and(
                mask, tf.logical_not(tf.logical_or(tmp_mask_inf, tmp_mask_sup))
            )
        return tf.cast(mask, "int32")

    def _single_slice(self, inputs):
        vol = inputs[0]
        crop_idx_inf = inputs[1]
        crop_idx_inf = tf.concat(
            [tf.cast(crop_idx_inf, "int32"), tf.zeros([1], dtype="int32")], axis=0
        )
        crop_size = tf.convert_to_tensor(
            [self.crop_shape] * self.n_dims + [-1], dtype="int32"
        )
        return tf.slice(vol, begin=crop_idx_inf, size=crop_size)

    def compute_output_shape(self, input_shape):
        return [
            (None, *[self.crop_shape] * self.n_dims, self.n_labels),
            (None, self.n_labels),
        ]



#!/usr/bin/env python3
"""
LaMAR: Label Augmented Modality Agnostic Registration
Command-line interface
"""

import argparse
import sys
import os
import tempfile
import shutil
from lamareg.scripts.lamar import lamareg
from lamareg.scripts import synthseg, coregister, apply_warp
from colorama import init, Fore, Style
import multiprocessing

init()

DEFAULT_THREADS = multiprocessing.cpu_count()


def print_cli_help():
    """Print a comprehensive help message for the LAMAR CLI."""
    # ANSI color codes
    CYAN = Fore.CYAN
    GREEN = Fore.GREEN
    YELLOW = Fore.YELLOW
    BLUE = Fore.BLUE
    MAGENTA = Fore.MAGENTA
    BOLD = Style.BRIGHT
    RESET = Style.RESET_ALL

    help_text = f"""
    {CYAN}{BOLD}╔════════════════════════════════════════════════════════════════╗
    ║                             LAMAR                              ║
    ║             Label Augmented Modality Agnostic Registration     ║
    ╚════════════════════════════════════════════════════════════════╝{RESET}

    LAMAR provides contrast-agnostic registration between different MRI modalities
    by using SynthSeg's brain parcellation to enable robust alignment between images
    with different contrasts (e.g., T1w to T2w, FLAIR to T1w, DWI to T1w).

    {CYAN}{BOLD}────────────────────────── WORKFLOWS ──────────────────────────{RESET}
    
    {BLUE}1. FULL REGISTRATION PIPELINE{RESET}
      Parcellate both input images, register them, and apply the transformation:
      lamar {GREEN}register{RESET} [options]
      
    {BLUE}2. GENERATE WARPFIELD ONLY{RESET}
      Create warpfields without applying them to the input image:
      lamar {GREEN}generate-warpfield{RESET} [options]
      
    {BLUE}3. APPLY EXISTING WARPFIELD{RESET}
      Apply previously created warpfields to an input image:
      lamar {GREEN}apply-warpfield{RESET} [options]
      
    {BLUE}4. DIRECT TOOL ACCESS{RESET}
      Run individual components directly:
      lamar {GREEN}synthseg{RESET} [options]     : Run SynthSeg brain parcellation
      lamar {GREEN}coregister{RESET} [options]   : Run ANTs coregistration
      lamar {GREEN}apply-warp{RESET} [options]   : Apply transformations
      lamar {GREEN}dice-compare{RESET} [options] : Calculate Dice similarity coefficient

    {CYAN}{BOLD}──────────────────── FULL REGISTRATION ────────────────────{RESET}
    
    {BLUE}# Required Arguments:{RESET}
      {YELLOW}--moving{RESET} PATH         : Input image to be registered
      {YELLOW}--fixed{RESET} PATH          : Reference image (target space)
      {YELLOW}--output{RESET} PATH         : Output registered image
      {YELLOW}--moving-parc{RESET} PATH    : Path for moving image parcellation
      {YELLOW}--fixed-parc{RESET} PATH     : Path for fixed image parcellation
      {YELLOW}--registered-parc{RESET} PATH: Path for registered parcellation
      {YELLOW}--affine{RESET} PATH         : Path for affine transformation
      {YELLOW}--warpfield{RESET} PATH      : Path for warp field
      
    {BLUE}# Optional Arguments:{RESET}
      {YELLOW}--registration-method{RESET} STR : Registration method (default: SyNRA)
      {YELLOW}--synthseg-threads{RESET} N      : SynthSeg threads (default: 1)
      {YELLOW}--ants-threads{RESET} N          : ANTs threads (default: 1)
      {YELLOW}--qc-csv{RESET} PATH             : Path for QC Dice score CSV file
      {YELLOW}--inverse-warpfield{RESET} PATH  : Path for inverse warp field
      {YELLOW}--inverse-affine{RESET} PATH     : Path for inverse affine transformation
      {YELLOW}--skip-fixed-parc{RESET}         : Toggle skipping fixed image parcellation
      {YELLOW}--skip-moving-parc{RESET}        : Toggle skipping moving image parcellation
      {YELLOW}--skip-qc{RESET}                 : Toggle skipping QC (default: False)
      {YELLOW}--disable-robust{RESET}          : Disable the two-stage robust registration (default: False)

    {CYAN}{BOLD}────────────────── GENERATE WARPFIELD ────────────────────{RESET}
    
    Same arguments as full registration, but without {YELLOW}--output{RESET}
    
    {CYAN}{BOLD}─────────────────── APPLY WARPFIELD ──────────────────────{RESET}
    
    {BLUE}# Required Arguments:{RESET}
      {YELLOW}--moving{RESET} PATH      : Input image to transform
      {YELLOW}--fixed{RESET} PATH       : Reference space image
      {YELLOW}--output{RESET} PATH      : Output registered image
      {YELLOW}--warpfield{RESET} PATH   : Path to warp field
      {YELLOW}--affine{RESET} PATH      : Path to affine transformation
      
    {BLUE}# Optional Arguments:{RESET}
      {YELLOW}--ants-threads{RESET} N   : ANTs threads (default: 1)

    {CYAN}{BOLD}─────────────────── EXAMPLE USAGE ───────────────────────{RESET}

    {BLUE}# Register DWI to T1w:{RESET}
    lamar {YELLOW}--moving{RESET} sub-001_dwi.nii.gz {YELLOW}--fixed{RESET} sub-001_T1w.nii.gz \\
      {YELLOW}--output{RESET} sub-001_dwi_in_T1w.nii.gz {YELLOW}--moving-parc{RESET} sub-001_dwi_parc.nii.gz \\
      {YELLOW}--fixed-parc{RESET} sub-001_T1w_parc.nii.gz {YELLOW}--registered-parc{RESET} sub-001_dwi_reg_parc.nii.gz \\
      {YELLOW}--affine{RESET} dwi_to_T1w_affine.mat {YELLOW}--warpfield{RESET} dwi_to_T1w_warp.nii.gz \\
      {YELLOW}--inverse-warpfield{RESET} T1w_to_dwi_warp.nii.gz {YELLOW}--inverse-affine{RESET} T1w_to_dwi_affine.mat \\
      {YELLOW}--synthseg-threads{RESET} 4 {YELLOW}--ants-threads{RESET} 8
      
    {BLUE}# Register without robust two-stage approach for challenging cases:{RESET}
    lamar {GREEN}register{RESET} {YELLOW}--moving{RESET} subject_flair.nii.gz {YELLOW}--fixed{RESET} subject_t1w.nii.gz \\
      {YELLOW}--output{RESET} registered_flair.nii.gz {YELLOW}--moving-parc{RESET} flair_parcellation.nii.gz \\
      {YELLOW}--fixed-parc{RESET} t1w_parcellation.nii.gz {YELLOW}--affine{RESET} flair_to_t1w_affine.mat \\
      {YELLOW}--warpfield{RESET} flair_to_t1w_warp.nii.gz {YELLOW}--disable-robust{RESET}


    {BLUE}# Generate parcellations separately:{RESET}
    lamar {GREEN}synthseg{RESET} {YELLOW}--i{RESET} subject_t1w.nii.gz {YELLOW}--o{RESET} t1w_parcellation.nii.gz {YELLOW}--parc{RESET}
    lamar {GREEN}synthseg{RESET} {YELLOW}--i{RESET} subject_flair.nii.gz {YELLOW}--o{RESET} flair_parcellation.nii.gz {YELLOW}--parc{RESET}

    {BLUE}# Register using existing parcellations:{RESET}
    lamar {GREEN}register{RESET} {YELLOW}--moving{RESET} subject_flair.nii.gz {YELLOW}--fixed{RESET} subject_t1w.nii.gz \\
      {YELLOW}--output{RESET} registered_flair.nii.gz {YELLOW}--moving-parc{RESET} flair_parcellation.nii.gz \\
      {YELLOW}--fixed-parc{RESET} t1w_parcellation.nii.gz {YELLOW}--skip-fixed-parc{RESET} {YELLOW}--skip-moving-parc{RESET} \\
      {YELLOW}--affine{RESET} flair_to_t1w_affine.mat {YELLOW}--warpfield{RESET} flair_to_t1w_warp.nii.gz

    {CYAN}{BOLD}────────────────────────── NOTES ───────────────────────{RESET}
    {MAGENTA}•{RESET} LAMAR works with any MRI modality combination
    {MAGENTA}•{RESET} If parcellation files already exist, they will be used directly
    {MAGENTA}•{RESET} All output files need explicit paths to ensure deterministic behavior
    {MAGENTA}•{RESET} The transforms can be reused with the apply-warpfield command
    {MAGENTA}•{RESET} Use dice-compare to evaluate registration quality
    {MAGENTA}•{RESET} The robust mode performs a two-stage registration for improved accuracy:
      1. Register parcellations (contrast-agnostic)
      2. Fine-tune with a second direct registration using the first result as initialization
    """
    print(help_text)


def main():
    """Main entry point for the LAMAR CLI."""
    parser = argparse.ArgumentParser(
        description="LAMAR: Label Augmented Modality Agnostic Registration"
    )

    # Create subparsers for different commands
    subparsers = parser.add_subparsers(dest="command", help="Command to run")
    # Check if we need to default to "register"
    if len(sys.argv) > 1 and sys.argv[1].startswith("-"):
        # The first argument is an option (starts with -), not a command
        # Insert "register" before the options
        new_argv = [sys.argv[0], "register"] + sys.argv[1:]
        print(f"No command specified, defaulting to 'register'")
        sys.argv = new_argv
        return main()  # Recursively call main with the modified arguments

    # WORKFLOW 1: Full registration pipeline
    register_parser = subparsers.add_parser(
        "register", help="Perform full registration pipeline with SynthSeg parcellation"
    )
    register_parser.add_argument(
        "--moving", required=True, help="Input moving image to be registered"
    )
    register_parser.add_argument(
        "--fixed", required=True, help="Reference fixed image (target space)"
    )
    register_parser.add_argument(
        "--skip-fixed-parc",
        action="store_true",
        help="Skip fixed image parcellation if it already exists",
    )
    register_parser.add_argument(
        "--skip-moving-parc",
        action="store_true",
        help="Skip moving image parcellation if it already exists",
    )
    register_parser.add_argument(
        "--output", required=True, help="Output registered image"
    )
    register_parser.add_argument(
        "--moving-parc",
        help="Output path for moving image parcellation (optional, will use temp file if not specified)",
    )
    register_parser.add_argument(
        "--fixed-parc",
        help="Output path for fixed image parcellation (optional, will use temp file if not specified)",
    )
    register_parser.add_argument(
        "--registered-parc",
        help="Output path for registered parcellation (optional, will use temp file if not specified)",
    )
    register_parser.add_argument(
        "--affine",
        help="Output path for affine transformation (optional, will use temp file if not specified)",
    )
    register_parser.add_argument(
        "--warpfield",
        help="Output path for warp field (optional, will use temp file if not specified)",
    )
    register_parser.add_argument(
        "--inverse-warpfield", help="Output path for inverse warp field (optional)"
    )
    register_parser.add_argument(
        "--inverse-affine",
        help="Output path for inverse affine transformation (optional)",
    )
    register_parser.add_argument(
        "--registration-method",
        default="SyNRA",
        help="Registration method (default: SyNRA)",
    )
    register_parser.add_argument(
        "--synthseg-threads",
        type=int,
        default=DEFAULT_THREADS,
        help="Number of threads to use for SynthSeg segmentation (default: 1)",
    )
    register_parser.add_argument(
        "--ants-threads",
        type=int,
        default=DEFAULT_THREADS,
        help="Number of threads to use for ANTs registration (default: 1)",
    )
    register_parser.add_argument(
        "--qc-csv", help="Path for quality control Dice score CSV file"
    )
    register_parser.add_argument(
        "--skip-qc", action="store_true", help="whether to skip QC (default: False)"
    )
    register_parser.add_argument(
        "--disable-robust",
        action="store_true",
        help="Whether to disable robust registration (default: False)",
    )

    # WORKFLOW 2: Generate warpfield only
    warpfield_parser = subparsers.add_parser(
        "generate-warpfield", help="Generate registration warpfield without applying it"
    )
    warpfield_parser.add_argument("--moving", required=True, help="Input moving image")
    warpfield_parser.add_argument(
        "--fixed", required=True, help="Reference fixed image"
    )
    warpfield_parser.add_argument(
        "--moving-parc",
        help="Output path for moving image parcellation (optional, will use temp file if not specified)",
    )
    warpfield_parser.add_argument(
        "--fixed-parc",
        help="Output path for fixed image parcellation (optional, will use temp file if not specified)",
    )
    warpfield_parser.add_argument(
        "--skip-fixed-parc",
        action="store_true",
        help="Skip fixed image parcellation if it already exists",
    )
    warpfield_parser.add_argument(
        "--skip-moving-parc",
        action="store_true",
        help="Skip moving image parcellation if it already exists",
    )
    warpfield_parser.add_argument(
        "--registered-parc",
        help="Output path for registered parcellation (optional, will use temp file if not specified)",
    )
    warpfield_parser.add_argument(
        "--affine",
        help="Output path for affine transformation (optional, will use temp file if not specified)",
    )
    warpfield_parser.add_argument(
        "--warpfield",
        help="Output path for warp field (optional, will use temp file if not specified)",
    )
    warpfield_parser.add_argument(
        "--inverse-warpfield", help="Output path for inverse warp field (optional)"
    )
    warpfield_parser.add_argument(
        "--inverse-affine",
        help="Output path for inverse affine transformation (optional)",
    )
    warpfield_parser.add_argument(
        "--registration-method",
        default="SyNRA",
        help="Registration method (default: SyNRA)",
    )
    warpfield_parser.add_argument(
        "--synthseg-threads",
        type=int,
        default=DEFAULT_THREADS,
        help="Number of threads to use for SynthSeg segmentation (default: 1)",
    )
    warpfield_parser.add_argument(
        "--ants-threads",
        type=int,
        default=DEFAULT_THREADS,
        help="Number of threads to use for ANTs registration (default: 1)",
    )
    warpfield_parser.add_argument(
        "--qc-csv", help="Path for quality control Dice score CSV file"
    )
    warpfield_parser.add_argument(
        "--skip-qc", action="store_true", help="whether to skip QC (default: False)"
    )
    warpfield_parser.add_argument(
        "--disable-robust",
        action="store_true",
        help="Whether to disable robust registration (default: False)",
    )

    # WORKFLOW 3: Apply existing warpfield
    apply_parser = subparsers.add_parser(
        "apply-warpfield", help="Apply existing warpfield to an image"
    )
    apply_parser.add_argument(
        "--moving", required=True, help="Input image to transform"
    )
    apply_parser.add_argument("--fixed", required=True, help="Reference space image")
    apply_parser.add_argument("--output", required=True, help="Output registered image")
    apply_parser.add_argument("--warpfield", required=True, help="Path to warp field")
    apply_parser.add_argument(
        "--affine", required=True, help="Path to affine transformation"
    )
    apply_parser.add_argument(
        "--ants-threads",
        type=int,
        default=DEFAULT_THREADS,
        help="Number of threads to use for ANTs transformation (default: 1)",
    )
    apply_parser.add_argument(
        "--inverse",
        action="store_true",
        help="Whether to invert the order of the affine and warpfield (warpfield first, then affine)"
    )
    
    # DIRECT TOOL ACCESS: SynthSeg
    synthseg_parser = subparsers.add_parser(
        "synthseg", help="Run SynthSeg brain MRI segmentation directly"
    )
    synthseg_parser.add_argument("--i", required=True, help="Input image")
    synthseg_parser.add_argument("--o", required=True, help="Output segmentation")
    synthseg_parser.add_argument(
        "--parc", action="store_true", help="Output parcellation"
    )
    synthseg_parser.add_argument("--cpu", action="store_true", help="Use CPU")
    synthseg_parser.add_argument(
        "--threads", type=int, default=DEFAULT_THREADS, help="Number of threads"
    )
    # Add other SynthSeg arguments as needed

    # DIRECT TOOL ACCESS: Coregister
    coregister_parser = subparsers.add_parser(
        "coregister", help="Run coregistration directly"
    )

    # DIRECT TOOL ACCESS: Apply Warp
    apply_warp_parser = subparsers.add_parser(
        "apply-warp", help="Apply transformation to an image directly"
    )

    # Add the dice-compare parser to the subparsers
    dice_compare_parser = subparsers.add_parser(
        "dice-compare",
        help="Calculate Dice similarity coefficient between two parcellation images",
    )
    dice_compare_parser.add_argument(
        "--ref", help="Path to reference parcellation image"
    )
    dice_compare_parser.add_argument(
        "--reg", help="Path to registered parcellation image"
    )
    dice_compare_parser.add_argument("--out", help="Output CSV file path")

    # Parse known args, leaving the rest for the subcommands
    args, unknown_args = parser.parse_known_args()

    # Print comprehensive help if no command provided
    if args.command is None:
        # Instead of just showing help, check if we have other arguments
        if len(sys.argv) > 1:
            # Insert "register" into the argument list and re-run
            new_argv = [sys.argv[0], "register"] + sys.argv[1:]
            print(f"No command specified, defaulting to 'register'")
            sys.argv = new_argv
            return main()  # Recursively call main with the new command
        else:
            # No arguments at all, show the help
            print_cli_help()
            return 0
    else:
        print(f"Command: {args.command}")

    # Handle command routing
    if args.command == "register":
        # Create a temporary directory for files that weren't specified
        temp_dir = None
        temp_files = []

        if not all(
            [
                args.moving_parc,
                args.fixed_parc,
                args.registered_parc,
                args.affine,
                args.warpfield,
            ]
        ):
            temp_dir = tempfile.mkdtemp(prefix="lamar_temp_")
            print(f"Created temporary directory for files: {temp_dir}")

        # Assign temporary paths for missing arguments
        if not args.moving_parc:
            args.moving_parc = os.path.join(temp_dir, "moving_parc.nii.gz")
            temp_files.append(args.moving_parc)

        if not args.fixed_parc:
            args.fixed_parc = os.path.join(temp_dir, "fixed_parc.nii.gz")
            temp_files.append(args.fixed_parc)

        if not args.registered_parc:
            args.registered_parc = os.path.join(temp_dir, "registered_parc.nii.gz")
            temp_files.append(args.registered_parc)

        if not args.affine:
            args.affine = os.path.join(temp_dir, "affine.mat")
            temp_files.append(args.affine)

        if not args.warpfield:
            args.warpfield = os.path.join(temp_dir, "warpfield.nii.gz")
            temp_files.append(args.warpfield)

        try:
            # Run the registration
            lamareg(
                input_image=args.moving,
                reference_image=args.fixed,
                output_image=args.output,
                input_parc=args.moving_parc,
                reference_parc=args.fixed_parc,
                output_parc=args.registered_parc,
                affine_file=args.affine,
                warp_file=args.warpfield,
                inverse_warp_file=args.inverse_warpfield,
                inverse_affine_file=args.inverse_affine,
                registration_method=args.registration_method,
                synthseg_threads=args.synthseg_threads,
                ants_threads=args.ants_threads,
                skip_fixed_parc=args.skip_fixed_parc,
                skip_moving_parc=args.skip_moving_parc,
                skip_qc=args.skip_qc,
                qc_csv=args.qc_csv,
                disable_robust=args.disable_robust,
            )

            # Clean up temporary files after successful completion
            if temp_dir:
                print(f"Cleaning up temporary files in {temp_dir}")
                shutil.rmtree(temp_dir)

        except Exception as e:
            print(f"Error during registration: {e}", file=sys.stderr)
            print(f"Temporary files remain in: {temp_dir}")
            sys.exit(1)
    elif args.command == "generate-warpfield":
        # Create a temporary directory for files that weren't specified
        temp_dir = None
        temp_files = []

        if not all(
            [
                args.moving_parc,
                args.fixed_parc,
                args.registered_parc,
                args.affine,
                args.warpfield,
            ]
        ):
            temp_dir = tempfile.mkdtemp(prefix="lamar_temp_")
            print(f"Created temporary directory for files: {temp_dir}")

        # Assign temporary paths for missing arguments
        if not args.moving_parc:
            args.moving_parc = os.path.join(temp_dir, "moving_parc.nii.gz")
            temp_files.append(args.moving_parc)

        if not args.fixed_parc:
            args.fixed_parc = os.path.join(temp_dir, "fixed_parc.nii.gz")
            temp_files.append(args.fixed_parc)

        if not args.registered_parc:
            args.registered_parc = os.path.join(temp_dir, "registered_parc.nii.gz")
            temp_files.append(args.registered_parc)

        if not args.affine:
            args.affine = os.path.join(temp_dir, "affine.mat")
            temp_files.append(args.affine)

        if not args.warpfield:
            args.warpfield = os.path.join(temp_dir, "warpfield.nii.gz")
            temp_files.append(args.warpfield)

        try:
            # Run the warpfield generation
            lamareg(
                input_image=args.moving,
                reference_image=args.fixed,
                output_image=None,  # No output image for generate-warpfield
                input_parc=args.moving_parc,
                reference_parc=args.fixed_parc,
                output_parc=args.registered_parc,
                affine_file=args.affine,
                warp_file=args.warpfield,
                inverse_warp_file=args.inverse_warpfield,
                inverse_affine_file=args.inverse_affine,
                generate_warpfield=True,
                registration_method=args.registration_method,
                synthseg_threads=args.synthseg_threads,
                ants_threads=args.ants_threads,
                skip_fixed_parc=args.skip_fixed_parc,
                skip_moving_parc=args.skip_moving_parc,
                skip_qc=args.skip_qc,
                qc_csv=args.qc_csv,
                disable_robust=args.disable_robust,
            )

            # Clean up temporary files after successful completion
            if temp_dir:
                print(f"Cleaning up temporary files in {temp_dir}")
                shutil.rmtree(temp_dir)

        except Exception as e:
            print(f"Error during warpfield generation: {e}", file=sys.stderr)
            if temp_dir:
                print(f"Temporary files remain in: {temp_dir}")
            sys.exit(1)
    elif args.command == "apply-warpfield":
        lamareg(
            input_image=args.moving,
            reference_image=args.fixed,
            output_image=args.output,
            apply_warpfield=True,
            affine_file=args.affine,
            warp_file=args.warpfield,
            ants_threads=args.ants_threads,
            inverse=args.inverse,
            synthseg_threads=1,  # Not used in this workflow but needed for the function
        )
    elif args.command == "synthseg":
        # Create a clean dictionary with the args provided by the parser
        synthseg_args = {}

        # Add explicit arguments from argparse
        if hasattr(args, "i") and args.i:
            synthseg_args["i"] = args.i
        if hasattr(args, "o") and args.o:
            synthseg_args["o"] = args.o

        # Add flag arguments
        for flag in ["parc", "cpu"]:
            if flag in unknown_args or f"--{flag}" in unknown_args:
                synthseg_args[flag] = True

        # Parse remaining arguments from command line
        i = 0
        while i < len(unknown_args):
            arg = unknown_args[i].lstrip("-")
            if i + 1 < len(unknown_args) and not unknown_args[i + 1].startswith("-"):
                synthseg_args[arg] = unknown_args[i + 1]
                i += 2
            else:
                # It's a flag
                synthseg_args[arg] = True
                i += 1

        # Set ALL required defaults for SynthSeg
        synthseg_args.setdefault("parc", True)
        synthseg_args.setdefault("cpu", True)
        synthseg_args.setdefault("robust", True)
        synthseg_args.setdefault("v1", False)
        synthseg_args.setdefault("fast", False)
        synthseg_args.setdefault("post", None)
        synthseg_args.setdefault("resample", None)
        synthseg_args.setdefault("ct", None)
        synthseg_args.setdefault("vol", None)
        synthseg_args.setdefault("qc", None)
        synthseg_args.setdefault("device", None)
        synthseg_args.setdefault("crop", None)

        if hasattr(args, "threads") and args.threads:
            synthseg_args["threads"] = str(args.threads)
        else:
            synthseg_args["threads"] = str(DEFAULT_THREADS)  # Use all available cores

        try:
            synthseg.main(synthseg_args)
        except Exception as e:
            print(f"SynthSeg error: {e}", file=sys.stderr)
            sys.exit(1)
    elif args.command == "coregister":
        # If no additional arguments are provided, print help
        if not unknown_args:
            coregister.print_help()
            sys.exit(0)
        # Forward arguments to coregister
        sys.argv = [sys.argv[0]] + unknown_args
        coregister.main()
    elif args.command == "apply-warp":
        # If no additional arguments are provided, print help
        if not unknown_args:
            apply_warp.print_help()
            sys.exit(0)
        # Forward arguments to apply_warp
        sys.argv = [sys.argv[0]] + unknown_args
        apply_warp.main()
    elif args.command == "dice-compare":
        from lamareg.scripts.dice_compare import compare_parcellations_dice, print_help

        print("Dice compare")
        if not hasattr(args, "ref") or not args.ref:
            print_help()
            sys.exit(0)

        compare_parcellations_dice(args.ref, args.reg, args.out)
    elif args.command is None:
        parser.print_help()
        sys.exit(0)
    else:
        print(f"Unknown command: {args.command}")
        parser.print_help()
        sys.exit(1)


if __name__ == "__main__":
    main()



"""

This function is very similar to training.py, as we train a UNet with synthetic data. In addition to the input image,
the UNet now also takes new inputs: soft probability maps for the target labels. These represent prior information, that
would typically be obtained at test time with a first segmenter.

If you use this code, please cite one of the SynthSeg papers:
https://github.com/BBillot/SynthSeg/blob/master/bibtex.bib

Copyright 2020 Benjamin Billot

Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software distributed under the License is
distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
implied. See the License for the specific language governing permissions and limitations under the
License.
"""

# python imports
import os
import numpy as np
import tensorflow as tf
from keras import models
import keras.layers as KL

# project imports
from lamareg.SynthSeg import metrics_model as metrics
from lamareg.SynthSeg.training import train_model
from lamareg.SynthSeg.brain_generator import BrainGenerator
from lamareg.SynthSeg.labels_to_image_model import get_shapes

# third-party imports
from lamareg.ext.lab2im import utils
from lamareg.ext.lab2im import layers
from lamareg.ext.neuron import models as nrn_models
from lamareg.ext.lab2im import edit_tensors as l2i_et
from lamareg.ext.lab2im.edit_volumes import get_ras_axes


def training(
    labels_dir,
    model_dir,
    generation_labels=None,
    grouping_labels=None,
    n_neutral_labels=None,
    segmentation_labels=None,
    subjects_prob=None,
    batchsize=1,
    n_channels=1,
    target_res=None,
    output_shape=None,
    generation_classes=None,
    prior_distributions="uniform",
    prior_means=None,
    prior_stds=None,
    use_specific_stats_for_channel=False,
    mix_prior_and_random=False,
    flipping=True,
    scaling_bounds=0.2,
    rotation_bounds=15,
    shearing_bounds=0.012,
    translation_bounds=False,
    nonlin_std=4.0,
    nonlin_scale=0.04,
    randomise_res=True,
    max_res_iso=4.0,
    max_res_aniso=8.0,
    data_res=None,
    thickness=None,
    bias_field_std=0.7,
    bias_scale=0.025,
    return_gradients=False,
    n_levels=5,
    nb_conv_per_level=2,
    conv_size=3,
    unet_feat_count=24,
    feat_multiplier=2,
    activation="elu",
    lr=1e-4,
    wl2_epochs=1,
    dice_epochs=50,
    steps_per_epoch=10000,
    checkpoint=None,
):
    """
    This function trains a UNet to segment MRI images with synthetic scans generated by sampling a GMM conditioned on
    label maps. The difference with training.py is based on the fact that here the UNet now takes two inputs:
    1) a (synthetic) image
    2) soft probability maps corresponding to a prior, which would typically be obtained at test time with a
    preliminary segmenter. Here, these probability maps are obtained from the training label maps, but undergo a
    degradation (i.e. spatial deformation and erosion/dilation), to model mistakes made by the preliminary segmenter at
    test time.

    We regroup the parameters in three categories: Generation, Architecture, Training.

    # IMPORTANT !!!
    # Each time we provide a parameter with separate values for each axis (e.g. with a numpy array or a sequence),
    # these values refer to the RAS axes.

    :param labels_dir: path of folder with all input label maps, or to a single label map (if only one training example)
    :param model_dir: path of a directory where the models will be saved during training.

    #---------------------------------------------- Generation parameters ----------------------------------------------
    # label maps parameters
    :param generation_labels: (optional) list of all possible label values in the input label maps.
    It can be None (default, where the label values are directly gotten from the provided label maps), a list,
    a 1d numpy array, or the path to such an array. If not None, the background label should always be the first. Then,
    if the label maps contain some right/left specific labels and if flipping is applied during training (see 'flipping'
    parameter), generation_labels should be organised as follows:
    first the background label, then the non-sided labels (i.e. those which are not right/left specific), then all the
    left labels, and finally the corresponding right labels (in the same order as the left ones). Please make sure each
    that each sided label has a right and a left value (this is essential!!!).
    :param n_neutral_labels: (optional) if the label maps contain some right/left specific labels and if flipping is
    applied during training, please provide the number of non-sided labels (including the background).
    This is used to know where the sided labels start in generation_labels. Leave to default (None) if either one of the
    two conditions is not fulfilled.
    :param segmentation_labels: (optional) list of the same length as generation_labels to indicate which values to use
    in the training label maps, i.e. all occurrences of generation_labels[i] in the input label maps will be converted
    to output_labels[i] in the returned label maps. Examples:
    Set output_labels[i] to zero if you wish to erase the value generation_labels[i] from the returned label maps.
    Set output_labels[i]=generation_labels[i] if you wish to keep the value generation_labels[i] in the returned maps.
    Can be a list or a 1d numpy array, or the path to such an array. Default is output_labels = generation_labels.
    :param grouping_labels: (optional) instead of simulating input soft probability maps for each target label, we can
    regroup label values in different groups. This is done with this parameter.
    Can be a list or a 1d numpy array, or the path to such an array.
    By default (None) soft probability maps are simulated for all segmentation labels.
    :param subjects_prob: (optional) relative order of importance (doesn't have to be probabilistic), with which to pick
    the provided label maps at each minibatch. Can be a sequence, a 1D numpy array, or the path to such an array, and it
    must be as long as path_label_maps. By default, all label maps are chosen with the same importance.

    # output-related parameters
    :param batchsize: (optional) number of images to generate per mini-batch. Default is 1.
    :param n_channels: (optional) number of channels to be synthesised. Default is 1.
    :param target_res: (optional) target resolution of the generated images and corresponding label maps.
    If None, the outputs will have the same resolution as the input label maps.
    Can be a number (isotropic resolution), or the path to a 1d numpy array.
    :param output_shape: (optional) desired shape of the output image, obtained by randomly cropping the generated image
    Can be an integer (same size in all dimensions), a sequence, a 1d numpy array, or the path to a 1d numpy array.
    Default is None, where no cropping is performed.

    # GMM-sampling parameters
    :param generation_classes: (optional) Indices regrouping generation labels into classes of same intensity
    distribution. Regrouped labels will thus share the same Gaussian when sampling a new image. Should be the path to a
    1d numpy array with the same length as generation_labels. and contain values between 0 and K-1, where K is the total
    number of classes. Default is all labels have different classes.
    Can be a list or a 1d numpy array, or the path to such an array.
    :param prior_distributions: (optional) type of distribution from which we sample the GMM parameters.
    Can either be 'uniform', or 'normal'. Default is 'uniform'.
    :param prior_means: (optional) hyperparameters controlling the prior distributions of the GMM means. Because
    these prior distributions are uniform or normal, they require by 2 hyperparameters. Can be a path to:
    1) an array of shape (2, K), where K is the number of classes (K=len(generation_labels) if generation_classes is
    not given). The mean of the Gaussian distribution associated to class k in [0, ...K-1] is sampled at each mini-batch
    from U(prior_means[0,k], prior_means[1,k]) if prior_distributions is uniform, and from
    N(prior_means[0,k], prior_means[1,k]) if prior_distributions is normal.
    2) an array of shape (2*n_mod, K), where each block of two rows is associated to hyperparameters derived
    from different modalities. In this case, if use_specific_stats_for_channel is False, we first randomly select a
    modality from the n_mod possibilities, and we sample the GMM means like in 2).
    If use_specific_stats_for_channel is True, each block of two rows correspond to a different channel
    (n_mod=n_channels), thus we select the corresponding block to each channel rather than randomly drawing it.
    Default is None, which corresponds all GMM means sampled from uniform distribution U(25, 225).
    :param prior_stds: (optional) same as prior_means but for the standard deviations of the GMM.
    Default is None, which corresponds to U(5, 25).
    :param use_specific_stats_for_channel: (optional) whether the i-th block of two rows in the prior arrays must be
    only used to generate the i-th channel. If True, n_mod should be equal to n_channels. Default is False.
    :param mix_prior_and_random: (optional) if prior_means is not None, enables to reset the priors to their default
    values for half of these cases, and thus generate images of random contrast.

    # spatial deformation parameters
    :param flipping: (optional) whether to introduce right/left random flipping. Default is True.
    :param scaling_bounds: (optional) if apply_linear_trans is True, the scaling factor for each dimension is
    sampled from a uniform distribution of predefined bounds. Can either be:
    1) a number, in which case the scaling factor is independently sampled from the uniform distribution of bounds
    (1-scaling_bounds, 1+scaling_bounds) for each dimension.
    2) the path to a numpy array of shape (2, n_dims), in which case the scaling factor in dimension i is sampled from
    the uniform distribution of bounds (scaling_bounds[0, i], scaling_bounds[1, i]) for the i-th dimension.
    3) False, in which case scaling is completely turned off.
    Default is scaling_bounds = 0.2 (case 1)
    :param rotation_bounds: (optional) same as scaling bounds but for the rotation angle, except that for case 1 the
    bounds are centred on 0 rather than 1, i.e. (0+rotation_bounds[i], 0-rotation_bounds[i]).
    Default is rotation_bounds = 15.
    :param shearing_bounds: (optional) same as scaling bounds. Default is shearing_bounds = 0.012.
    :param translation_bounds: (optional) same as scaling bounds. Default is translation_bounds = False, but we
    encourage using it when cropping is deactivated (i.e. when output_shape=None).
    :param nonlin_std: (optional) Standard deviation of the normal distribution from which we sample the first
    tensor for synthesising the deformation field. Set to 0 to completely deactivate elastic deformation.
    :param nonlin_scale: (optional) Ratio between the size of the input label maps and the size of the sampled
    tensor for synthesising the elastic deformation field.

    # blurring/resampling parameters
    :param randomise_res: (optional) whether to mimic images that would have been 1) acquired at low resolution, and
    2) resampled to high resolution. The low resolution is uniformly resampled at each minibatch from [1mm, 9mm].
    In that process, the images generated by sampling the GMM are: 1) blurred at the sampled LR, 2) downsampled at LR,
    and 3) resampled at target_resolution.
    :param max_res_iso: (optional) If randomise_res is True, this enables to control the upper bound of the uniform
    distribution from which we sample the random resolution U(min_res, max_res_iso), where min_res is the resolution of
    the input label maps. Must be a number, and default is 4. Set to None to deactivate it, but if randomise_res is
    True, at least one of max_res_iso or max_res_aniso must be given.
    :param max_res_aniso: If randomise_res is True, this enables to downsample the input volumes to a random LR in
    only 1 (random) direction. This is done by randomly selecting a direction i in the range [0, n_dims-1], and sampling
    a value in the corresponding uniform distribution U(min_res[i], max_res_aniso[i]), where min_res is the resolution
    of the input label maps. Can be a number, a sequence, or a 1d numpy array. Set to None to deactivate it, but if
    randomise_res is True, at least one of max_res_iso or max_res_aniso must be given.
    :param data_res: (optional) specific acquisition resolution to mimic, as opposed to random resolution sampled when
    randomise_res is True. This triggers a blurring which mimics the acquisition resolution, but downsampling is
    optional (see param downsample). Default for data_res is None, where images are slightly blurred. If the generated
    images are uni-modal, data_res can be a number (isotropic acquisition resolution), a sequence, a 1d numpy array, or
    the path to a 1d numpy array. In the multi-modal case, it should be given as a numpy array (or a path) of size
    (n_mod, n_dims), where each row is the acquisition resolution of the corresponding channel.
    :param thickness: (optional) if data_res is provided, we can further specify the slice thickness of the low
    resolution images to mimic. Must be provided in the same format as data_res. Default thickness = data_res.

    # bias field parameters
    :param bias_field_std: (optional) If strictly positive, this triggers the corruption of images with a bias field.
    The bias field is obtained by sampling a first small tensor from a normal distribution, resizing it to
    full size, and rescaling it to positive values by taking the voxel-wise exponential. bias_field_std designates the
    std dev of the normal distribution from which we sample the first tensor.
    Set to 0 to completely deactivate bias field corruption.
    :param bias_scale: (optional) If bias_field_std is not False, this designates the ratio between the size of
    the input label maps and the size of the first sampled tensor for synthesising the bias field.

    :param return_gradients: (optional) whether to return the synthetic image or the magnitude of its spatial gradient
    (computed with Sobel kernels).

    # ------------------------------------------ UNet architecture parameters ------------------------------------------
    :param n_levels: (optional) number of level for the Unet. Default is 5.
    :param nb_conv_per_level: (optional) number of convolutional layers per level. Default is 2.
    :param conv_size: (optional) size of the convolution kernels. Default is 2.
    :param unet_feat_count: (optional) number of feature for the first layer of the UNet. Default is 24.
    :param feat_multiplier: (optional) multiply the number of feature by this number at each new level. Default is 2.
    :param activation: (optional) activation function. Can be 'elu', 'relu'.

    # ----------------------------------------------- Training parameters ----------------------------------------------
    :param lr: (optional) learning rate for the training. Default is 1e-4
    :param wl2_epochs: (optional) number of epochs for which the network (except the soft-max layer) is trained with L2
    norm loss function. Default is 1.
    :param dice_epochs: (optional) number of epochs with the soft Dice loss function. Default is 50.
    :param steps_per_epoch: (optional) number of steps per epoch. Default is 10000. Since no online validation is
    possible, this is equivalent to the frequency at which the models are saved.
    :param checkpoint: (optional) path of an already saved model to load before starting the training.
    """

    # check epochs
    assert (wl2_epochs > 0) | (
        dice_epochs > 0
    ), "either wl2_epochs or dice_epochs must be positive, had {0} and {1}".format(
        wl2_epochs, dice_epochs
    )

    # get label lists
    generation_labels, _ = utils.get_list_labels(
        label_list=generation_labels, labels_dir=labels_dir
    )
    if segmentation_labels is not None:
        segmentation_labels, _ = utils.get_list_labels(label_list=segmentation_labels)
    else:
        segmentation_labels = generation_labels
    n_segmentation_labels = len(np.unique(segmentation_labels))

    # instantiate BrainGenerator object
    brain_generator = BrainGeneratorGroup(
        labels_dir=labels_dir,
        generation_labels=generation_labels,
        grouping_labels=grouping_labels,
        n_neutral_labels=n_neutral_labels,
        output_labels=segmentation_labels,
        subjects_prob=subjects_prob,
        batchsize=batchsize,
        n_channels=n_channels,
        target_res=target_res,
        output_shape=output_shape,
        output_div_by_n=2**n_levels,
        generation_classes=generation_classes,
        prior_distributions=prior_distributions,
        prior_means=prior_means,
        prior_stds=prior_stds,
        use_specific_stats_for_channel=use_specific_stats_for_channel,
        mix_prior_and_random=mix_prior_and_random,
        flipping=flipping,
        scaling_bounds=scaling_bounds,
        rotation_bounds=rotation_bounds,
        shearing_bounds=shearing_bounds,
        translation_bounds=translation_bounds,
        nonlin_std=nonlin_std,
        nonlin_scale=nonlin_scale,
        randomise_res=randomise_res,
        max_res_iso=max_res_iso,
        max_res_aniso=max_res_aniso,
        data_res=data_res,
        thickness=thickness,
        bias_field_std=bias_field_std,
        bias_scale=bias_scale,
        return_gradients=return_gradients,
    )

    # generation model
    labels_to_image_model = brain_generator.labels_to_image_model
    unet_input_shape = brain_generator.model_output_shape

    # prepare the segmentation model
    unet_model = nrn_models.unet(
        input_model=labels_to_image_model,
        input_shape=unet_input_shape,
        nb_labels=n_segmentation_labels,
        nb_levels=n_levels,
        nb_conv_per_level=nb_conv_per_level,
        conv_size=conv_size,
        nb_features=unet_feat_count,
        feat_mult=feat_multiplier,
        activation=activation,
        batch_norm=-1,
        name="unet2",
    )

    # input generator
    input_generator = utils.build_training_generator(
        brain_generator.model_inputs_generator, batchsize
    )

    # pre-training with weighted L2, input is fit to the softmax rather than the probabilities
    if wl2_epochs > 0:
        wl2_model = models.Model(
            unet_model.inputs, [unet_model.get_layer("unet2_likelihood").output]
        )
        wl2_model = metrics.metrics_model(wl2_model, segmentation_labels, "wl2")
        train_model(
            wl2_model,
            input_generator,
            lr,
            wl2_epochs,
            steps_per_epoch,
            model_dir,
            "wl2",
            checkpoint,
        )
        checkpoint = os.path.join(model_dir, "wl2_%03d.h5" % wl2_epochs)

    # fine-tuning with dice metric
    dice_model = metrics.metrics_model(unet_model, segmentation_labels, "dice")
    train_model(
        dice_model,
        input_generator,
        lr,
        dice_epochs,
        steps_per_epoch,
        model_dir,
        "dice",
        checkpoint,
    )


class BrainGeneratorGroup(BrainGenerator):

    def __init__(self, grouping_labels=None, **kwargs):
        super(BrainGeneratorGroup, self).__init__(**kwargs)
        self.grouping_labels = utils.load_array_if_path(grouping_labels)
        self.labels_to_image_model, self.model_output_shape = (
            self._build_labels_to_image_model_group()
        )

    def _build_labels_to_image_model_group(self):
        # build_model
        lab_to_im_model = labels_to_image_model_group(
            labels_shape=self.labels_shape,
            n_channels=self.n_channels,
            generation_labels=self.generation_labels,
            output_labels=self.output_labels,
            n_neutral_labels=self.n_neutral_labels,
            atlas_res=self.atlas_res,
            target_res=self.target_res,
            grouping_labels=self.grouping_labels,
            output_shape=self.output_shape,
            output_div_by_n=self.output_div_by_n,
            flipping=self.flipping,
            aff=np.eye(4),
            scaling_bounds=self.scaling_bounds,
            rotation_bounds=self.rotation_bounds,
            shearing_bounds=self.shearing_bounds,
            translation_bounds=self.translation_bounds,
            nonlin_std=self.nonlin_std,
            nonlin_scale=self.nonlin_scale,
            randomise_res=self.randomise_res,
            max_res_iso=self.max_res_iso,
            max_res_aniso=self.max_res_aniso,
            data_res=self.data_res,
            thickness=self.thickness,
            bias_field_std=self.bias_field_std,
            bias_scale=self.bias_scale,
            return_gradients=self.return_gradients,
        )
        out_shape = lab_to_im_model.output[0].get_shape().as_list()[1:]
        return lab_to_im_model, out_shape


def labels_to_image_model_group(
    labels_shape,
    n_channels,
    generation_labels,
    output_labels,
    n_neutral_labels,
    atlas_res,
    target_res,
    grouping_labels,
    output_shape=None,
    output_div_by_n=None,
    flipping=True,
    aff=None,
    scaling_bounds=0.2,
    rotation_bounds=15,
    shearing_bounds=0.012,
    translation_bounds=False,
    nonlin_std=3.0,
    nonlin_scale=0.0625,
    randomise_res=False,
    max_res_iso=4.0,
    max_res_aniso=8.0,
    data_res=None,
    thickness=None,
    bias_field_std=0.7,
    bias_scale=0.025,
    return_gradients=False,
):

    # reformat resolutions
    labels_shape = utils.reformat_to_list(labels_shape)
    n_dims, _ = utils.get_dims(labels_shape)
    atlas_res = utils.reformat_to_n_channels_array(atlas_res, n_dims, n_channels)
    data_res = (
        atlas_res
        if data_res is None
        else utils.reformat_to_n_channels_array(data_res, n_dims, n_channels)
    )
    thickness = (
        data_res
        if thickness is None
        else utils.reformat_to_n_channels_array(thickness, n_dims, n_channels)
    )
    atlas_res = atlas_res[0]
    target_res = (
        atlas_res
        if target_res is None
        else utils.reformat_to_n_channels_array(target_res, n_dims)[0]
    )

    # get shapes
    crop_shape, output_shape = get_shapes(
        labels_shape, output_shape, atlas_res, target_res, output_div_by_n
    )

    # define model inputs
    labels_input = KL.Input(
        shape=labels_shape + [1], name="labels_input", dtype="int32"
    )
    means_input = KL.Input(
        shape=list(generation_labels.shape) + [n_channels], name="means_input"
    )
    stds_input = KL.Input(
        shape=list(generation_labels.shape) + [n_channels], name="std_devs_input"
    )
    list_inputs = [labels_input, means_input, stds_input]

    # deform labels
    labels = layers.RandomSpatialDeformation(
        scaling_bounds=scaling_bounds,
        rotation_bounds=rotation_bounds,
        shearing_bounds=shearing_bounds,
        translation_bounds=translation_bounds,
        nonlin_std=nonlin_std,
        nonlin_scale=nonlin_scale,
        inter_method="nearest",
    )(labels_input)

    # get mask and further deforms/dilates it
    grouped_labels = layers.ConvertLabels(
        generation_labels, dest_values=grouping_labels
    )(labels)
    grouped_labels = layers.RandomSpatialDeformation(
        scaling_bounds=0.05,
        rotation_bounds=3,
        shearing_bounds=0.007,
        translation_bounds=2,
        nonlin_std=1.5,
        nonlin_scale=0.04,
        inter_method="nearest",
        prob_deform=0.9,
    )(grouped_labels)

    # randomly dilate/erode binary mask of each group
    n_group = len(np.unique(grouping_labels))
    grouped_labels = KL.Lambda(
        lambda x: tf.one_hot(tf.cast(x[..., 0], "int32"), n_group, axis=-1)
    )(grouped_labels)
    split = KL.Lambda(lambda x: tf.split(x, [1] * n_group, axis=-1))(grouped_labels)
    channels = [
        layers.RandomDilationErosion(
            1, 1, 5, 0.9, operation="random", return_mask=True
        )(c)
        for c in split
    ]
    grouped_labels = KL.Lambda(lambda x: tf.concat(x, -1), name="group_morph")(channels)

    # cropping
    if crop_shape != labels_shape:
        labels, grouped_labels = layers.RandomCrop(crop_shape)([labels, grouped_labels])

    # flipping
    if flipping:
        assert aff is not None, "aff should not be None if flipping is True"
        labels, grouped_labels = layers.RandomFlip(
            get_ras_axes(aff, n_dims)[0],
            [True, False],
            generation_labels,
            n_neutral_labels,
        )([labels, grouped_labels])

    # build synthetic image
    image = layers.SampleConditionalGMM(generation_labels)(
        [labels, means_input, stds_input]
    )

    # apply bias field
    if bias_field_std > 0:
        image = layers.BiasFieldCorruption(bias_field_std, bias_scale, False)(image)

    # intensity augmentation
    image = layers.IntensityAugmentation(
        clip=300, normalise=True, gamma_std=0.5, separate_channels=True
    )(image)

    # loop over channels
    channels = list()
    split = (
        KL.Lambda(lambda x: tf.split(x, [1] * n_channels, axis=-1))(image)
        if (n_channels > 1)
        else [image]
    )
    for i, channel in enumerate(split):

        if randomise_res:
            max_res_iso = np.array(
                utils.reformat_to_list(max_res_iso, length=n_dims, dtype="float")
            )
            max_res_aniso = np.array(
                utils.reformat_to_list(max_res_aniso, length=n_dims, dtype="float")
            )
            max_res = np.maximum(max_res_iso, max_res_aniso)
            resolution, blur_res = layers.SampleResolution(
                atlas_res, max_res_iso, max_res_aniso
            )(means_input)
            sigma = l2i_et.blurring_sigma_for_downsampling(
                atlas_res, resolution, thickness=blur_res
            )
            channel = layers.DynamicGaussianBlur(
                0.75 * max_res / np.array(atlas_res), 1.03
            )([channel, sigma])
            channel = layers.MimicAcquisition(
                atlas_res, atlas_res, output_shape, False
            )([channel, resolution])
            channels.append(channel)

        else:
            sigma = l2i_et.blurring_sigma_for_downsampling(
                atlas_res, data_res[i], thickness=thickness[i]
            )
            channel = layers.GaussianBlur(sigma, 1.03)(channel)
            resolution = KL.Lambda(
                lambda x: tf.convert_to_tensor(data_res[i], dtype="float32")
            )([])
            channel = layers.MimicAcquisition(atlas_res, data_res[i], output_shape)(
                [channel, resolution]
            )
            channels.append(channel)

    # concatenate all channels back
    image = (
        KL.Lambda(lambda x: tf.concat(x, -1))(channels)
        if len(channels) > 1
        else channels[0]
    )

    # compute image gradient
    if return_gradients:
        image = layers.ImageGradients("sobel", True, name="image_gradients")(image)
        image = layers.IntensityAugmentation(clip=10, normalise=True)(image)

    # resample labels at target resolution
    if crop_shape != output_shape:
        labels = l2i_et.resample_tensor(labels, output_shape, interp_method="nearest")
        grouped_labels = l2i_et.resample_tensor(
            grouped_labels, output_shape, interp_method="nearest"
        )

    # map generation labels to segmentation values
    labels = layers.ConvertLabels(
        generation_labels, dest_values=output_labels, name="labels_out"
    )(labels)

    # build model (dummy layer enables to keep the labels when plugging this model to other models)
    image = KL.Lambda(
        lambda x: tf.concat([x[0], tf.cast(x[1], dtype=x[0].dtype)], axis=-1),
        name="image_out",
    )([image, grouped_labels, labels])
    brain_model = models.Model(inputs=list_inputs, outputs=[image, labels])

    return brain_model
